# === UCSF ChimeraX Copyright ===
# Copyright 2016 Regents of the University of California.
# All rights reserved.  This software provided pursuant to a
# license agreement containing restrictions on its disclosure,
# duplication and use.  For details see:
# http://www.rbvi.ucsf.edu/chimerax/docs/licensing.html
# This notice must be embedded in or attached to all copies,
# including partial copies, of the software or any revisions
# or derivations thereof.
# === UCSF ChimeraX Copyright ===

# -----------------------------------------------------------------------------
# Provide symmetry operators for any of 230 space groups using PDB naming
# convention.
#

# -----------------------------------------------------------------------------
#
sg_symop_table = None
def space_group_symmetry_table():

    global sg_symop_table
    if sg_symop_table == None:
        sg_symop_table = {}
        symops = parse_symop_data(space_group_symmetry_operations)
        for name, ops in symops:
            sg_symop_table[name] = ops
        for alt_name, name in alternate_names:
            sg_symop_table[alt_name] = sg_symop_table[name]
            
    return sg_symop_table

# -----------------------------------------------------------------------------
# Ordered by space group number (1-230) in the international tables.
#
sg_230 = []
def standard_groups_230():

    global sg_230
    if len(sg_230) == 0:
        sg_230 = parse_symop_data(ebi_autodep_symmetry_operations)
    return sg_230


# -----------------------------------------------------------------------------
#
def parse_symop_data(symops_string):

    lines = symops_string.split('\n')

    symops = []
    name = None
    for line in lines:
        if len(line) == 0:      # count newline
            if name and len(ops) != count:
                raise SyntaxError('Wrong number of symmetry operators for ' + name)
            continue
        elif line[0] == '(':
            symop = parse_symop(line)
            ops.append(symop)
        else:
            fields = line.split()
            count = int(fields[-1])
            name = ' '.join(fields[:-1])
            ops = []
            symops.append((name, ops))

    return symops

# -----------------------------------------------------------------------------
#
def parse_symop(line):

    sline = line.strip()
    if sline and sline[0] == '(':
        sline = sline[1:]
    if sline and sline[-1] == ')':
        sline = sline[:-1]
    fields = sline.split(',')
    if len(fields) != 3:
        raise SyntaxError('Bad symmetry operator format: ' + line)

    tf = [parse_symop_axis(fields[k]) for k in range(3)]
    from chimerax.geometry import Place
    p = Place(tf)

    return p

# -----------------------------------------------------------------------------
#
def parse_symop_axis(string):

    rt = [0,0,0,0]
    
    axes = (('+X', 0, 1), ('-X', 0, -1), ('X', 0, 1), ('+x', 0, 1), ('-x', 0, -1), ('x', 0, 1),
            ('+Y', 1, 1), ('-Y', 1, -1), ('Y', 1, 1), ('+y', 1, 1), ('-y', 1, -1), ('y', 1, 1),
            ('+Z', 2, 1), ('-Z', 2, -1), ('Z', 2, 1), ('+z', 2, 1), ('-z', 2, -1), ('z', 2, 1))
    for name, axis, sign in axes:
        if string.find(name) >= 0:
            string = string.replace(name, '')
            rt[axis] = sign

    if string:
        fields = string.split('/')
        if len(fields) != 2:
            raise SyntaxError('Bad symmetry operator component: ' + string)
        rt[3] = float(fields[0]) / float(fields[1])

    return rt

# -----------------------------------------------------------------------------
# Some PDB files use alternate space group names.  By comparing the SMTRY
# remarks in such files, the preferred names in autodep_symop.txt from the
# EBI PDB submission site can be found.
#
alternate_names = (
    ('C 1 2 1', 'C 2'),
    ('P 1 2 1', 'P 2'),
    ('P 1 21 1', 'P 21'),
)

# -----------------------------------------------------------------------------
# This table of the 230 space groups comes from the EBI PDB deposition web site
#
#       http://autodep.ebi.ac.uk/sgtable.html
#
# Each space group was in a separate file so I wrote the fetch_symop.py script
# to download them all from the web.  That script made the file version
# autodep_symop.txt.
#
# Renamed "R 3" to "H 3" and "R 3 2" to "H 3 2".  The R names were not the correct
# Hermann Mauguin symbols for the listed symmetry operators, e.g. PDB 1CTJ.
#
ebi_autodep_symmetry_operations = \
"""
P 1	1
(X,Y,Z)

P -1	2
(X,Y,Z)
(-X,-Y,-Z)

P 2	2
(X,Y,Z)
(-X,Y,-Z)

P 21	2
(X,Y,Z)
(-X,Y+1/2,-Z)

C 2	4
(X,Y,Z)
(-X,Y,-Z)
(1/2+X,1/2+Y,Z)
(1/2-X,1/2+Y,-Z)

P M	2
(X,Y,Z)
(X,-Y,Z)

P C	2
(X,Y,Z)
(X,-Y,1/2+Z)

C M	4
(X,Y,Z)
(X,-Y,Z)
(1/2+X,1/2+Y,Z)
(1/2+X,1/2-Y,Z)

C C	4
(X,Y,Z)
(X,-Y,1/2+Z)
(1/2+X,1/2+Y,Z)
(1/2+X,1/2-Y,1/2+Z)

P 2/M	4
(X,Y,Z)
(X,-Y,Z)
(-X,Y,-Z)
(-X,-Y,-Z)

P 21/M	4
(X,Y,Z)
(-X,1/2+Y,-Z)
(-X,-Y,-Z)
(X,1/2-Y,Z)

C 2/M	8
(X,Y,Z)
(X,-Y,Z)
(-X,Y,-Z)
(-X,-Y,-Z)
(1/2+X,1/2+Y,Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,-Z)
(1/2-X,1/2-Y,-Z)

P 2/C	4
(X,Y,Z)
(-X,Y,1/2-Z)
(-X,-Y,-Z)
(X,-Y,1/2+Z)

P 21/C	4
(X,Y,Z)
(-X,-Y,-Z)
(-X,1/2+Y,1/2-Z)
(X,1/2-Y,1/2+Z)

C 2/C	8
(X,Y,Z)
(-X,Y,1/2-Z)
(-X,-Y,-Z)
(X,-Y,1/2+Z)
(1/2+X,1/2+Y,Z)
(1/2-X,1/2+Y,1/2-Z)
(1/2-X,1/2-Y,-Z)
(1/2+X,1/2-Y,1/2+Z)

P 2 2 2	4
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)

P 2 2 21	4
(X,Y,Z)
(-X,-Y,1/2+Z)
(-X,Y,1/2-Z)
(X,-Y,-Z)

P 21 21 2	4
(X,Y,Z)
(-X,-Y,Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,1/2-Y,-Z)

P 21 21 21	4
(X,Y,Z)
(1/2-X,-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,-Z)

C 2 2 21	8
(X,Y,Z)
(-X,-Y,1/2+Z)
(-X,Y,1/2-Z)
(X,-Y,-Z)
(1/2+X,1/2+Y,Z)
(1/2-X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,-Z)

C 2 2 2	8
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(1/2+X,1/2+Y,Z)
(1/2-X,1/2-Y,Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,1/2-Y,-Z)

F 2 2 2	16
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(X,1/2+Y,1/2+Z)
(-X,1/2-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(X,1/2-Y,1/2-Z)
(1/2+X,Y,1/2+Z)
(1/2-X,-Y,1/2+Z)
(1/2-X,Y,1/2-Z)
(1/2+X,-Y,1/2-Z)
(1/2+X,1/2+Y,Z)
(1/2-X,1/2-Y,Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,1/2-Y,-Z)

I 2 2 2	8
(X,Y,Z)
(-X,-Y,Z)
(X,-Y,-Z)
(-X,Y,-Z)
(X+1/2,Y+1/2,Z+1/2)
(-X+1/2,-Y+1/2,Z+1/2)
(X+1/2,-Y+1/2,-Z+1/2)
(-X+1/2,Y+1/2,-Z+1/2)

I 21 21 21	8
(X,Y,Z)
(1/2-X,-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,-Z)
(1/2+X,1/2+Y,1/2+Z)
(-X,1/2-Y,Z)
(1/2-X,Y,-Z)
(X,-Y,1/2-Z)

P M M 2	4
(X,Y,Z)
(-X,-Y,Z)
(X,-Y,Z)
(-X,Y,Z)

P M C 21	4
(X,Y,Z)
(-X,-Y,1/2+Z)
(X,-Y,1/2+Z)
(-X,Y,Z)

P C C 2	4
(X,Y,Z)
(-X,-Y,Z)
(X,-Y,1/2+Z)
(-X,Y,1/2+Z)

P M A 2	4
(X,Y,Z)
(-X,-Y,Z)
(1/2+X,-Y,Z)
(1/2-X,Y,Z)

P C A 21	4
(X,Y,Z)
(-X,-Y,1/2+Z)
(1/2+X,-Y,Z)
(1/2-X,Y,1/2+Z)

P N C 2	4
(X,Y,Z)
(-X,-Y,Z)
(X,1/2-Y,1/2+Z)
(-X,1/2+Y,1/2+Z)

P M N 21	4
(X,Y,Z)
(1/2-X,-Y,1/2+Z)
(1/2+X,-Y,1/2+Z)
(-X,Y,Z)

P B A 2	4
(X,Y,Z)
(-X,-Y,Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,Z)

P N A 21	4
(X,Y,Z)
(-X,-Y,1/2+Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,1/2+Z)

P N N 2	4
(X,Y,Z)
(-X,-Y,Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2+Z)

C M M 2	8
(X,Y,Z)
(-X,-Y,Z)
(X,-Y,Z)
(-X,Y,Z)
(1/2+X,1/2+Y,Z)
(1/2-X,1/2-Y,Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,Z)

C M C 21	8
(X,Y,Z)
(-X,-Y,1/2+Z)
(X,-Y,1/2+Z)
(-X,Y,Z)
(1/2+X,1/2+Y,Z)
(1/2-X,1/2-Y,1/2+Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,Z)

C C C 2	8
(X,Y,Z)
(-X,-Y,Z)
(X,-Y,1/2+Z)
(-X,Y,1/2+Z)
(1/2+X,1/2+Y,Z)
(1/2-X,1/2-Y,Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2+Z)

A M M 2	8
(X,Y,Z)
(-X,-Y,Z)
(X,-Y,Z)
(-X,Y,Z)
(X,1/2+Y,1/2+Z)
(-X,1/2-Y,1/2+Z)
(X,1/2-Y,1/2+Z)
(-X,1/2+Y,1/2+Z)

A B M 2	8
(X,Y,Z)
(-X,-Y,Z)
(X,1/2-Y,Z)
(-X,1/2+Y,Z)
(X,1/2+Y,1/2+Z)
(-X,1/2-Y,1/2+Z)
(X,-Y,1/2+Z)
(-X,Y,1/2+Z)

A M A 2	8
(X,Y,Z)
(-X,-Y,Z)
(1/2+X,-Y,Z)
(1/2-X,Y,Z)
(X,1/2+Y,1/2+Z)
(-X,1/2-Y,1/2+Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2+Z)

A B A 2	8
(X,Y,Z)
(-X,-Y,Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,Z)
(X,1/2+Y,1/2+Z)
(-X,1/2-Y,1/2+Z)
(1/2+X,-Y,1/2+Z)
(1/2-X,Y,1/2+Z)

F M M 2	16
(X,Y,Z)
(-X,-Y,Z)
(X,-Y,Z)
(-X,Y,Z)
(X,1/2+Y,1/2+Z)
(-X,1/2-Y,1/2+Z)
(X,1/2-Y,1/2+Z)
(-X,1/2+Y,1/2+Z)
(1/2+X,Y,1/2+Z)
(1/2-X,-Y,1/2+Z)
(1/2+X,-Y,1/2+Z)
(1/2-X,Y,1/2+Z)
(1/2+X,1/2+Y,Z)
(1/2-X,1/2-Y,Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,Z)

F D D 2	16
(X,Y,Z)
(-X,-Y,Z)
(1/4+X,1/4-Y,1/4+Z)
(1/4-X,1/4+Y,1/4+Z)
(X,1/2+Y,1/2+Z)
(-X,1/2-Y,1/2+Z)
(1/4+X,3/4-Y,3/4+Z)
(1/4-X,3/4+Y,3/4+Z)
(1/2+X,Y,1/2+Z)
(1/2-X,-Y,1/2+Z)
(3/4+X,1/4-Y,3/4+Z)
(3/4-X,1/4+Y,3/4+Z)
(1/2+X,1/2+Y,Z)
(1/2-X,1/2-Y,Z)
(3/4+X,3/4-Y,1/4+Z)
(3/4-X,3/4+Y,1/4+Z)

I M M 2	8
(X,Y,Z)
(-X,-Y,Z)
(X,-Y,Z)
(-X,Y,Z)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,1/2-Y,1/2+Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2+Z)

I B A 2	8
(X,Y,Z)
(-X,-Y,Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,Z)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,1/2-Y,1/2+Z)
(X,-Y,1/2+Z)
(-X,Y,1/2+Z)

I M A 2	8
(X,Y,Z)
(-X,-Y,Z)
(1/2+X,-Y,Z)
(1/2-X,Y,Z)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,1/2-Y,1/2+Z)
(X,1/2-Y,1/2+Z)
(-X,1/2+Y,1/2+Z)

P M M M	8
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(-X,-Y,-Z)
(X,Y,-Z)
(X,-Y,Z)
(-X,Y,Z)

P N N N	8
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(1/2-X,1/2-Y,1/2-Z)
(1/2+X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2+Z)

P C C M	8
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,1/2-Z)
(X,-Y,1/2-Z)
(-X,-Y,-Z)
(X,Y,-Z)
(X,-Y,1/2+Z)
(-X,Y,1/2+Z)

P B A N	8
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(1/2-X,1/2-Y,-Z)
(1/2+X,1/2+Y,-Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,Z)

P M M A	8
(X,Y,Z)
(1/2-X,-Y,Z)
(-X,Y,-Z)
(1/2+X,-Y,-Z)
(-X,-Y,-Z)
(1/2+X,Y,-Z)
(X,-Y,Z)
(1/2-X,Y,Z)

P N N A	8
(X,Y,Z)
(1/2-X,-Y,Z)
(1/2-X,1/2+Y,1/2-Z)
(X,1/2-Y,1/2-Z)
(-X,-Y,-Z)
(1/2+X,Y,-Z)
(1/2+X,1/2-Y,1/2+Z)
(-X,1/2+Y,1/2+Z)

P M N A	8
(X,Y,Z)
(1/2-X,-Y,1/2+Z)
(1/2-X,Y,1/2-Z)
(X,-Y,-Z)
(-X,-Y,-Z)
(1/2+X,Y,1/2-Z)
(1/2+X,-Y,1/2+Z)
(-X,Y,Z)

P C C A	8
(X,Y,Z)
(1/2-X,-Y,Z)
(-X,Y,1/2-Z)
(1/2+X,-Y,1/2-Z)
(-X,-Y,-Z)
(1/2+X,Y,-Z)
(X,-Y,1/2+Z)
(1/2-X,Y,1/2+Z)

P B A M	8
(X,Y,Z)
(-X,-Y,Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,1/2-Y,-Z)
(-X,-Y,-Z)
(X,Y,-Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,Z)

P C C N	8
(X,Y,Z)
(1/2-X,1/2-Y,Z)
(-X,1/2+Y,1/2-Z)
(1/2+X,-Y,1/2-Z)
(-X,-Y,-Z)
(1/2+X,1/2+Y,-Z)
(X,1/2-Y,1/2+Z)
(1/2-X,Y,1/2+Z)

P B C M	8
(X,Y,Z)
(-X,-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(X,1/2-Y,-Z)
(-X,-Y,-Z)
(X,Y,1/2-Z)
(X,1/2-Y,1/2+Z)
(-X,1/2+Y,Z)

P N N M	8
(X,Y,Z)
(-X,-Y,Z)
(1/2-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2-Z)
(-X,-Y,-Z)
(X,Y,-Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2+Z)

P M M N	8
(X,Y,Z)
(-X,-Y,Z)
(1/2-X,Y+1/2,-Z)
(X+1/2,1/2-Y,-Z)
(1/2-X,1/2-Y,-Z)
(X+1/2,Y+1/2,-Z)
(X,-Y,Z)
(-X,Y,Z)

P B C N	8
(X,Y,Z)
(1/2-X,1/2-Y,1/2+Z)
(-X,Y,1/2-Z)
(1/2+X,1/2-Y,-Z)
(-X,-Y,-Z)
(1/2+X,1/2+Y,1/2-Z)
(X,-Y,1/2+Z)
(1/2-X,1/2+Y,Z)

P B C A	8
(X,Y,Z)
(1/2-X,-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,-Z)
(-X,-Y,-Z)
(1/2+X,Y,1/2-Z)
(X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,Z)

P N M A	8
(X,Y,Z)
(-X+1/2,-Y,Z+1/2)
(-X,Y+1/2,-Z)
(X+1/2,-Y+1/2,-Z+1/2)
(-X,-Y,-Z)
(X+1/2,Y,-Z+1/2)
(X,-Y+1/2,Z)
(-X+1/2,Y+1/2,Z+1/2)

C M C M	16
(X,Y,Z)
(-X,-Y,1/2+Z)
(-X,Y,1/2-Z)
(X,-Y,-Z)
(-X,-Y,-Z)
(X,Y,1/2-Z)
(X,-Y,1/2+Z)
(-X,Y,Z)
(1/2+X,1/2+Y,Z)
(1/2-X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,-Z)
(1/2-X,1/2-Y,-Z)
(1/2+X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,Z)

C M C A	16
(X,Y,Z)
(-X,1/2-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(X,-Y,-Z)
(-X,-Y,-Z)
(X,1/2+Y,1/2-Z)
(X,1/2-Y,1/2+Z)
(-X,Y,Z)
(1/2+X,1/2+Y,Z)
(1/2-X,-Y,1/2+Z)
(1/2-X,Y,1/2-Z)
(1/2+X,1/2-Y,-Z)
(1/2-X,1/2-Y,-Z)
(1/2+X,Y,1/2-Z)
(1/2+X,-Y,1/2+Z)
(1/2-X,1/2+Y,Z)

C M M M	16
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(-X,-Y,-Z)
(X,Y,-Z)
(X,-Y,Z)
(-X,Y,Z)
(1/2+X,1/2+Y,Z)
(1/2-X,1/2-Y,Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,1/2-Y,-Z)
(1/2-X,1/2-Y,-Z)
(1/2+X,1/2+Y,-Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,Z)

C C C M	16
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,1/2-Z)
(X,-Y,1/2-Z)
(-X,-Y,-Z)
(X,Y,-Z)
(X,-Y,1/2+Z)
(-X,Y,1/2+Z)
(1/2+X,1/2+Y,Z)
(1/2-X,1/2-Y,Z)
(1/2-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2-Z)
(1/2-X,1/2-Y,-Z)
(1/2+X,1/2+Y,-Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2+Z)

C M M A	16
(X,Y,Z)
(-X,1/2-Y,Z)
(-X,1/2+Y,-Z)
(X,-Y,-Z)
(-X,-Y,-Z)
(X,1/2+Y,-Z)
(X,1/2-Y,Z)
(-X,Y,Z)
(1/2+X,1/2+Y,Z)
(1/2-X,-Y,Z)
(1/2-X,Y,-Z)
(1/2+X,1/2-Y,-Z)
(1/2-X,1/2-Y,-Z)
(1/2+X,Y,-Z)
(1/2+X,-Y,Z)
(1/2-X,1/2+Y,Z)

C C C A	16
(X,Y,Z)
(1/2-X,1/2-Y,Z)
(-X,Y,-Z)
(1/2+X,1/2-Y,-Z)
(-X,1/2-Y,1/2-Z)
(1/2+X,Y,1/2-Z)
(X,1/2-Y,1/2+Z)
(1/2-X,Y,1/2+Z)
(1/2+X,1/2+Y,Z)
(-X,-Y,Z)
(1/2-X,1/2+Y,-Z)
(X,-Y,-Z)
(1/2-X,-Y,1/2-Z)
(X,1/2+Y,1/2-Z)
(1/2+X,-Y,1/2+Z)
(-X,1/2+Y,1/2+Z)

F M M M	32
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(-X,-Y,-Z)
(X,Y,-Z)
(X,-Y,Z)
(-X,Y,Z)
(X,1/2+Y,1/2+Z)
(-X,1/2-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(X,1/2-Y,1/2-Z)
(-X,1/2-Y,1/2-Z)
(X,1/2+Y,1/2-Z)
(X,1/2-Y,1/2+Z)
(-X,1/2+Y,1/2+Z)
(1/2+X,Y,1/2+Z)
(1/2-X,-Y,1/2+Z)
(1/2-X,Y,1/2-Z)
(1/2+X,-Y,1/2-Z)
(1/2-X,-Y,1/2-Z)
(1/2+X,Y,1/2-Z)
(1/2+X,-Y,1/2+Z)
(1/2-X,Y,1/2+Z)
(1/2+X,1/2+Y,Z)
(1/2-X,1/2-Y,Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,1/2-Y,-Z)
(1/2-X,1/2-Y,-Z)
(1/2+X,1/2+Y,-Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,Z)

F D D D	32
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(1/4-X,1/4-Y,1/4-Z)
(1/4+X,1/4+Y,1/4-Z)
(1/4+X,1/4-Y,1/4+Z)
(1/4-X,1/4+Y,1/4+Z)
(X,1/2+Y,1/2+Z)
(-X,1/2-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(X,1/2-Y,1/2-Z)
(1/4-X,3/4-Y,3/4-Z)
(1/4+X,3/4+Y,3/4-Z)
(1/4+X,3/4-Y,3/4+Z)
(1/4-X,3/4+Y,3/4+Z)
(1/2+X,Y,1/2+Z)
(1/2-X,-Y,1/2+Z)
(1/2-X,Y,1/2-Z)
(1/2+X,-Y,1/2-Z)
(3/4-X,1/4-Y,3/4-Z)
(3/4+X,1/4+Y,3/4-Z)
(3/4+X,1/4-Y,3/4+Z)
(3/4-X,1/4+Y,3/4+Z)
(1/2+X,1/2+Y,Z)
(1/2-X,1/2-Y,Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,1/2-Y,-Z)
(3/4-X,3/4-Y,1/4-Z)
(3/4+X,3/4+Y,1/4-Z)
(3/4+X,3/4-Y,1/4+Z)
(3/4-X,3/4+Y,1/4+Z)

I M M M	16
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(-X,-Y,-Z)
(X,Y,-Z)
(X,-Y,Z)
(-X,Y,Z)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2-Z)
(1/2-X,1/2-Y,1/2-Z)
(1/2+X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2+Z)

I B A M	16
(X,Y,Z)
(-X,-Y,Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,1/2-Y,-Z)
(-X,-Y,-Z)
(X,Y,-Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,Z)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,1/2-Y,1/2+Z)
(-X,Y,1/2-Z)
(X,-Y,1/2-Z)
(1/2-X,1/2-Y,1/2-Z)
(1/2+X,1/2+Y,1/2-Z)
(X,-Y,1/2+Z)
(-X,Y,1/2+Z)

I B C A	16
(X,Y,Z)
(1/2-X,-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,-Z)
(-X,-Y,-Z)
(1/2+X,Y,1/2-Z)
(X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,Z)
(1/2+X,1/2+Y,1/2+Z)
(-X,1/2-Y,Z)
(1/2-X,Y,-Z)
(X,-Y,1/2-Z)
(1/2-X,1/2-Y,1/2-Z)
(X,1/2+Y,-Z)
(1/2+X,-Y,Z)
(-X,Y,1/2+Z)

I M M A	16
(X,Y,Z)
(-X,1/2-Y,Z)
(-X,1/2+Y,-Z)
(X,-Y,-Z)
(-X,-Y,-Z)
(X,1/2+Y,-Z)
(X,1/2-Y,Z)
(-X,Y,Z)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,-Y,1/2+Z)
(1/2-X,Y,1/2-Z)
(1/2+X,1/2-Y,1/2-Z)
(1/2-X,1/2-Y,1/2-Z)
(1/2+X,Y,1/2-Z)
(1/2+X,-Y,1/2+Z)
(1/2-X,1/2+Y,1/2+Z)

P 4	4
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,Z)
(Y,-X,Z)

P 41	4
(X,Y,Z)
(-X,-Y,1/2+Z)
(-Y,X,1/4+Z)
(Y,-X,3/4+Z)

P 42	4
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,1/2+Z)
(Y,-X,1/2+Z)

P 43	4
(X,Y,Z)
(-X,-Y,1/2+Z)
(-Y,X,3/4+Z)
(Y,-X,1/4+Z)

I 4	8
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,Z)
(Y,-X,Z)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,1/2-Y,1/2+Z)
(1/2-Y,1/2+X,1/2+Z)
(1/2+Y,1/2-X,1/2+Z)

I 41	8
(X,Y,Z)
(1/2-X,1/2-Y,1/2+Z)
(-Y,1/2+X,1/4+Z)
(1/2+Y,-X,3/4+Z)
(1/2+X,1/2+Y,1/2+Z)
(-X,-Y,Z)
(1/2-Y,X,3/4+Z)
(Y,1/2-X,1/4+Z)

P -4	4
(X,Y,Z)
(-X,-Y,Z)
(Y,-X,-Z)
(-Y,X,-Z)

I -4	8
(X,Y,Z)
(-X,-Y,Z)
(Y,-X,-Z)
(-Y,X,-Z)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,1/2-Y,1/2+Z)
(1/2+Y,1/2-X,1/2-Z)
(1/2-Y,1/2+X,1/2-Z)

P 4/M	8
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,Z)
(Y,-X,Z)
(-X,-Y,-Z)
(X,Y,-Z)
(Y,-X,-Z)
(-Y,X,-Z)

P 42/M	8
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,1/2+Z)
(Y,-X,1/2+Z)
(-X,-Y,-Z)
(X,Y,-Z)
(Y,-X,1/2-Z)
(-Y,X,1/2-Z)

P 4/N	8
(X,Y,Z)
(-X,-Y,Z)
(1/2-Y,1/2+X,Z)
(1/2+Y,1/2-X,Z)
(1/2-X,1/2-Y,-Z)
(1/2+X,1/2+Y,-Z)
(Y,-X,-Z)
(-Y,X,-Z)

P 42/N	8
(X,Y,Z)
(-X,-Y,Z)
(1/2-Y,1/2+X,1/2+Z)
(1/2+Y,1/2-X,1/2+Z)
(1/2-X,1/2-Y,1/2-Z)
(1/2+X,1/2+Y,1/2-Z)
(Y,-X,-Z)
(-Y,X,-Z)

I 4/M	16
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,Z)
(Y,-X,Z)
(-X,-Y,-Z)
(X,Y,-Z)
(Y,-X,-Z)
(-Y,X,-Z)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,1/2-Y,1/2+Z)
(1/2-Y,1/2+X,1/2+Z)
(1/2+Y,1/2-X,1/2+Z)
(1/2-X,1/2-Y,1/2-Z)
(1/2+X,1/2+Y,1/2-Z)
(1/2+Y,1/2-X,1/2-Z)
(1/2-Y,1/2+X,1/2-Z)

I 41/A	16
(X,Y,Z)
(1/2-X,1/2-Y,1/2+Z)
(-Y,1/2+X,1/4+Z)
(1/2+Y,-X,3/4+Z)
(-X,1/2-Y,1/4-Z)
(1/2+X,Y,3/4-Z)
(Y,-X,-Z)
(1/2-Y,1/2+X,1/2-Z)
(1/2+X,1/2+Y,1/2+Z)
(-X,-Y,Z)
(1/2-Y,X,3/4+Z)
(Y,1/2-X,1/4+Z)
(1/2-X,-Y,3/4-Z)
(X,1/2+Y,1/4-Z)
(1/2+Y,1/2-X,1/2-Z)
(-Y,X,-Z)

P 4 2 2	8
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,Z)
(Y,-X,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Y,X,-Z)
(-Y,-X,-Z)

P 4 21 2	8
(X,Y,Z)
(-X,-Y,Z)
(1/2-Y,1/2+X,Z)
(1/2+Y,1/2-X,Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,1/2-Y,-Z)
(Y,X,-Z)
(-Y,-X,-Z)

P 41 2 2	8
(X,Y,Z)
(-X,-Y,1/2+Z)
(-Y,X,1/4+Z)
(Y,-X,3/4+Z)
(-X,Y,-Z)
(X,-Y,1/2-Z)
(Y,X,3/4-Z)
(-Y,-X,1/4-Z)

P 41 21 2	8
(X,Y,Z)
(-X,-Y,1/2+Z)
(1/2-Y,1/2+X,1/4+Z)
(1/2+Y,1/2-X,3/4+Z)
(1/2-X,1/2+Y,1/4-Z)
(1/2+X,1/2-Y,3/4-Z)
(Y,X,-Z)
(-Y,-X,1/2-Z)

P 42 2 2	8
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,1/2+Z)
(Y,-X,1/2+Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Y,X,1/2-Z)
(-Y,-X,1/2-Z)

P 42 21 2	8
(X,Y,Z)
(-X,-Y,Z)
(1/2-Y,1/2+X,1/2+Z)
(1/2+Y,1/2-X,1/2+Z)
(1/2-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2-Z)
(Y,X,-Z)
(-Y,-X,-Z)

P 43 2 2	8
(X,Y,Z)
(-X,-Y,1/2+Z)
(-Y,X,3/4+Z)
(Y,-X,1/4+Z)
(-X,Y,-Z)
(X,-Y,1/2-Z)
(Y,X,1/4-Z)
(-Y,-X,3/4-Z)

P 43 21 2	8
(X,Y,Z)
(-X,-Y,1/2+Z)
(1/2-Y,1/2+X,3/4+Z)
(1/2+Y,1/2-X,1/4+Z)
(1/2-X,1/2+Y,3/4-Z)
(1/2+X,1/2-Y,1/4-Z)
(Y,X,-Z)
(-Y,-X,1/2-Z)

I 4 2 2	16
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,Z)
(Y,-X,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Y,X,-Z)
(-Y,-X,-Z)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,1/2-Y,1/2+Z)
(1/2-Y,1/2+X,1/2+Z)
(1/2+Y,1/2-X,1/2+Z)
(1/2-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2-Z)
(1/2+Y,1/2+X,1/2-Z)
(1/2-Y,1/2-X,1/2-Z)

I 41 2 2	16
(X,Y,Z)
(1/2-X,1/2-Y,1/2+Z)
(-Y,1/2+X,1/4+Z)
(1/2+Y,-X,3/4+Z)
(1/2-X,Y,3/4-Z)
(X,1/2-Y,1/4-Z)
(1/2+Y,1/2+X,1/2-Z)
(-Y,-X,-Z)
(1/2+X,1/2+Y,1/2+Z)
(-X,-Y,Z)
(1/2-Y,X,3/4+Z)
(Y,1/2-X,1/4+Z)
(-X,1/2+Y,1/4-Z)
(1/2+X,-Y,3/4-Z)
(Y,X,-Z)
(1/2-Y,1/2-X,1/2-Z)

P 4 M M	8
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,Z)
(Y,-X,Z)
(X,-Y,Z)
(-X,Y,Z)
(-Y,-X,Z)
(Y,X,Z)

P 4 B M	8
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,Z)
(Y,-X,Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,Z)
(1/2-Y,1/2-X,Z)
(1/2+Y,1/2+X,Z)

P 42 C M	8
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,1/2+Z)
(Y,-X,1/2+Z)
(X,-Y,1/2+Z)
(-X,Y,1/2+Z)
(-Y,-X,Z)
(Y,X,Z)

P 42 N M	8
(X,Y,Z)
(-X,-Y,Z)
(1/2-Y,1/2+X,1/2+Z)
(1/2+Y,1/2-X,1/2+Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2+Z)
(-Y,-X,Z)
(Y,X,Z)

P 4 C C	8
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,Z)
(Y,-X,Z)
(X,-Y,1/2+Z)
(-X,Y,1/2+Z)
(-Y,-X,1/2+Z)
(Y,X,1/2+Z)

P 4 N C	8
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,Z)
(Y,-X,Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2+Z)
(1/2-Y,1/2-X,1/2+Z)
(1/2+Y,1/2+X,1/2+Z)

P 42 M C	8
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,1/2+Z)
(Y,-X,1/2+Z)
(X,-Y,Z)
(-X,Y,Z)
(-Y,-X,1/2+Z)
(Y,X,1/2+Z)

P 42 B C	8
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,1/2+Z)
(Y,-X,1/2+Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,Z)
(1/2-Y,1/2-X,1/2+Z)
(1/2+Y,1/2+X,1/2+Z)

I 4 M M	16
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,Z)
(Y,-X,Z)
(X,-Y,Z)
(-X,Y,Z)
(-Y,-X,Z)
(Y,X,Z)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,1/2-Y,1/2+Z)
(1/2-Y,1/2+X,1/2+Z)
(1/2+Y,1/2-X,1/2+Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2+Z)
(1/2-Y,1/2-X,1/2+Z)
(1/2+Y,1/2+X,1/2+Z)

I 4 C M	16
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,Z)
(Y,-X,Z)
(X,-Y,1/2+Z)
(-X,Y,1/2+Z)
(-Y,-X,1/2+Z)
(Y,X,1/2+Z)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,1/2-Y,1/2+Z)
(1/2-Y,1/2+X,1/2+Z)
(1/2+Y,1/2-X,1/2+Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,Z)
(1/2-Y,1/2-X,Z)
(1/2+Y,1/2+X,Z)

I 41 M D	16
(X,Y,Z)
(1/2-X,1/2-Y,1/2+Z)
(-Y,1/2+X,1/4+Z)
(1/2+Y,-X,3/4+Z)
(X,-Y,Z)
(1/2-X,1/2+Y,1/2+Z)
(-Y,1/2-X,1/4+Z)
(1/2+Y,X,3/4+Z)
(1/2+X,1/2+Y,1/2+Z)
(-X,-Y,Z)
(1/2-Y,X,3/4+Z)
(Y,1/2-X,1/4+Z)
(1/2+X,1/2-Y,1/2+Z)
(-X,Y,Z)
(1/2-Y,-X,3/4+Z)
(Y,1/2+X,1/4+Z)

I 41 C D	16
(X,Y,Z)
(1/2-X,1/2-Y,1/2+Z)
(-Y,1/2+X,1/4+Z)
(1/2+Y,-X,3/4+Z)
(X,-Y,1/2+Z)
(1/2-X,1/2+Y,Z)
(-Y,1/2-X,3/4+Z)
(1/2+Y,X,1/4+Z)
(1/2+X,1/2+Y,1/2+Z)
(-X,-Y,Z)
(1/2-Y,X,3/4+Z)
(Y,1/2-X,1/4+Z)
(1/2+X,1/2-Y,Z)
(-X,Y,1/2+Z)
(1/2-Y,-X,1/4+Z)
(Y,1/2+X,3/4+Z)

P -4 2 M	8
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,-Z)
(Y,-X,-Z)
(-X,Y,-Z)
(X,-Y,-Z)
(-Y,-X,Z)
(Y,X,Z)

P -4 2 C	8
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,-Z)
(Y,-X,-Z)
(-X,Y,1/2-Z)
(X,-Y,1/2-Z)
(-Y,-X,1/2+Z)
(Y,X,1/2+Z)

P -4 21 M	8
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,-Z)
(Y,-X,-Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,1/2-Y,-Z)
(1/2-Y,1/2-X,Z)
(1/2+Y,1/2+X,Z)

P -4 21 C	8
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,-Z)
(Y,-X,-Z)
(1/2-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2-Z)
(1/2-Y,1/2-X,1/2+Z)
(1/2+Y,1/2+X,1/2+Z)

P -4 M 2	8
(X,Y,Z)
(-X,-Y,Z)
(Y,-X,-Z)
(-Y,X,-Z)
(X,-Y,Z)
(-X,Y,Z)
(Y,X,-Z)
(-Y,-X,-Z)

P -4 C 2	8
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,-Z)
(Y,-X,-Z)
(X,-Y,1/2+Z)
(-X,Y,1/2+Z)
(Y,X,1/2-Z)
(-Y,-X,1/2-Z)

P -4 B 2	8
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,-Z)
(Y,-X,-Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,Z)
(1/2+Y,1/2+X,-Z)
(1/2-Y,1/2-X,-Z)

P -4 N 2	8
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,-Z)
(Y,-X,-Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2+Z)
(1/2+Y,1/2+X,1/2-Z)
(1/2-Y,1/2-X,1/2-Z)

I -4 M 2	16
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,-Z)
(Y,-X,-Z)
(X,-Y,Z)
(-X,Y,Z)
(Y,X,-Z)
(-Y,-X,-Z)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,1/2-Y,1/2+Z)
(1/2-Y,1/2+X,1/2-Z)
(1/2+Y,1/2-X,1/2-Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2+Z)
(1/2+Y,1/2+X,1/2-Z)
(1/2-Y,1/2-X,1/2-Z)

I -4 C 2	16
(X,Y,Z)
(-X,-Y,Z)
(Y,-X,-Z)
(-Y,X,-Z)
(X,-Y,1/2+Z)
(-X,Y,1/2+Z)
(Y,X,1/2-Z)
(-Y,-X,1/2-Z)
(X+1/2,Y+1/2,Z+1/2)
(-X+1/2,-Y+1/2,Z+1/2)
(Y+1/2,-X+1/2,-Z+1/2)
(-Y+1/2,X+1/2,-Z+1/2)
(X+1/2,-Y+1/2,Z)
(-X+1/2,Y+1/2,Z)
(Y+1/2,X+1/2,-Z)
(-Y+1/2,-X+1/2,-Z)

I -4 2 M	16
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,-Z)
(Y,-X,-Z)
(-X,Y,-Z)
(X,-Y,-Z)
(-Y,-X,Z)
(Y,X,Z)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,1/2-Y,1/2+Z)
(1/2-Y,1/2+X,1/2-Z)
(1/2+Y,1/2-X,1/2-Z)
(1/2-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2-Z)
(1/2-Y,1/2-X,1/2+Z)
(1/2+Y,1/2+X,1/2+Z)

I -4 2 D	16
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,-Z)
(Y,-X,-Z)
(1/2-X,Y,3/4-Z)
(1/2+X,-Y,3/4-Z)
(1/2-Y,-X,3/4+Z)
(1/2+Y,X,3/4+Z)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,1/2-Y,1/2+Z)
(1/2-Y,1/2+X,1/2-Z)
(1/2+Y,1/2-X,1/2-Z)
(-X,1/2+Y,1/4-Z)
(X,1/2-Y,1/4-Z)
(-Y,1/2-X,1/4+Z)
(Y,1/2+X,1/4+Z)

P 4/M M M	16
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,Z)
(Y,-X,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Y,X,-Z)
(-Y,-X,-Z)
(-X,-Y,-Z)
(X,Y,-Z)
(Y,-X,-Z)
(-Y,X,-Z)
(X,-Y,Z)
(-X,Y,Z)
(-Y,-X,Z)
(Y,X,Z)

P 4/M C C	16
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,Z)
(Y,-X,Z)
(-X,Y,1/2-Z)
(X,-Y,1/2-Z)
(Y,X,1/2-Z)
(-Y,-X,1/2-Z)
(-X,-Y,-Z)
(X,Y,-Z)
(Y,-X,-Z)
(-Y,X,-Z)
(X,-Y,1/2+Z)
(-X,Y,1/2+Z)
(-Y,-X,1/2+Z)
(Y,X,1/2+Z)

P 4/N B M	16
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,Z)
(Y,-X,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Y,X,-Z)
(-Y,-X,-Z)
(1/2-X,1/2-Y,-Z)
(1/2+X,1/2+Y,-Z)
(1/2+Y,1/2-X,-Z)
(1/2-Y,1/2+X,-Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,Z)
(1/2-Y,1/2-X,Z)
(1/2+Y,1/2+X,Z)

P 4/N N C	16
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,Z)
(Y,-X,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Y,X,-Z)
(-Y,-X,-Z)
(1/2-X,1/2-Y,1/2-Z)
(1/2+X,1/2+Y,1/2-Z)
(1/2+Y,1/2-X,1/2-Z)
(1/2-Y,1/2+X,1/2-Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2+Z)
(1/2-Y,1/2-X,1/2+Z)
(1/2+Y,1/2+X,1/2+Z)

P 4/M B M	16
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,Z)
(Y,-X,Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,1/2-Y,-Z)
(1/2+Y,1/2+X,-Z)
(1/2-Y,1/2-X,-Z)
(-X,-Y,-Z)
(X,Y,-Z)
(Y,-X,-Z)
(-Y,X,-Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,Z)
(1/2-Y,1/2-X,Z)
(1/2+Y,1/2+X,Z)

P 4/M N C	16
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,Z)
(Y,-X,Z)
(1/2-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2-Z)
(1/2+Y,1/2+X,1/2-Z)
(1/2-Y,1/2-X,1/2-Z)
(-X,-Y,-Z)
(X,Y,-Z)
(Y,-X,-Z)
(-Y,X,-Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2+Z)
(1/2-Y,1/2-X,1/2+Z)
(1/2+Y,1/2+X,1/2+Z)

P 4/N M M	16
(X,Y,Z)
(-X,-Y,Z)
(1/2-Y,1/2+X,Z)
(1/2+Y,1/2-X,Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,1/2-Y,-Z)
(Y,X,-Z)
(-Y,-X,-Z)
(1/2-X,1/2-Y,-Z)
(1/2+X,1/2+Y,-Z)
(Y,-X,-Z)
(-Y,X,-Z)
(X,-Y,Z)
(-X,Y,Z)
(1/2-Y,1/2-X,Z)
(1/2+Y,1/2+X,Z)

P 4/N C C	16
(X,Y,Z)
(-X,-Y,Z)
(1/2-Y,1/2+X,Z)
(1/2+Y,1/2-X,Z)
(1/2-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2-Z)
(Y,X,1/2-Z)
(-Y,-X,1/2-Z)
(1/2-X,1/2-Y,-Z)
(1/2+X,1/2+Y,-Z)
(Y,-X,-Z)
(-Y,X,-Z)
(X,-Y,1/2+Z)
(-X,Y,1/2+Z)
(1/2-Y,1/2-X,1/2+Z)
(1/2+Y,1/2+X,1/2+Z)

P 42/M M C	16
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,1/2+Z)
(Y,-X,1/2+Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Y,X,1/2-Z)
(-Y,-X,1/2-Z)
(-X,-Y,-Z)
(X,Y,-Z)
(Y,-X,1/2-Z)
(-Y,X,1/2-Z)
(X,-Y,Z)
(-X,Y,Z)
(-Y,-X,1/2+Z)
(Y,X,1/2+Z)

P 42/M C M	16
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,1/2+Z)
(Y,-X,1/2+Z)
(-X,Y,1/2-Z)
(X,-Y,1/2-Z)
(Y,X,-Z)
(-Y,-X,-Z)
(-X,-Y,-Z)
(X,Y,-Z)
(Y,-X,1/2-Z)
(-Y,X,1/2-Z)
(X,-Y,1/2+Z)
(-X,Y,1/2+Z)
(-Y,-X,Z)
(Y,X,Z)

P 42/N B C	16
(X,Y,Z)
(-X,-Y,Z)
(1/2-Y,1/2+X,1/2+Z)
(1/2+Y,1/2-X,1/2+Z)
(-X,Y,1/2-Z)
(X,-Y,1/2-Z)
(1/2+Y,1/2+X,-Z)
(1/2-Y,1/2-X,-Z)
(1/2-X,1/2-Y,1/2-Z)
(1/2+X,1/2+Y,1/2-Z)
(Y,-X,-Z)
(-Y,X,-Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,Z)
(-Y,-X,1/2+Z)
(Y,X,1/2+Z)

P 42/N N M	16
(X,Y,Z)
(-X,-Y,Z)
(1/2-Y,1/2+X,1/2+Z)
(1/2+Y,1/2-X,1/2+Z)
(-X,Y,-Z)
(X,-Y,-Z)
(1/2+Y,1/2+X,1/2-Z)
(1/2-Y,1/2-X,1/2-Z)
(1/2-X,1/2-Y,1/2-Z)
(1/2+X,1/2+Y,1/2-Z)
(Y,-X,-Z)
(-Y,X,-Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2+Z)
(-Y,-X,Z)
(Y,X,Z)

P 42/M B C	16
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,1/2+Z)
(Y,-X,1/2+Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,1/2-Y,-Z)
(1/2+Y,1/2+X,1/2-Z)
(1/2-Y,1/2-X,1/2-Z)
(-X,-Y,-Z)
(X,Y,-Z)
(Y,-X,1/2-Z)
(-Y,X,1/2-Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,Z)
(1/2-Y,1/2-X,1/2+Z)
(1/2+Y,1/2+X,1/2+Z)

P 42/M N M	16
(X,Y,Z)
(-X,-Y,Z)
(1/2-Y,X+1/2,Z+1/2)
(Y+1/2,1/2-X,Z+1/2)
(1/2-X,Y+1/2,1/2-Z)
(X+1/2,1/2-Y,1/2-Z)
(Y,X,-Z)
(-Y,-X,-Z)
(-X,-Y,-Z)
(X,Y,-Z)
(Y+1/2,1/2-X,1/2-Z)
(1/2-Y,X+1/2,1/2-Z)
(X+1/2,1/2-Y,Z+1/2)
(1/2-X,Y+1/2,Z+1/2)
(-Y,-X,Z)
(Y,X,Z)

P 42/N M C	16
(X,Y,Z)
(-X,-Y,Z)
(1/2-Y,1/2+X,1/2+Z)
(1/2+Y,1/2-X,1/2+Z)
(1/2-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2-Z)
(Y,X,-Z)
(-Y,-X,-Z)
(1/2-X,1/2-Y,1/2-Z)
(1/2+X,1/2+Y,1/2-Z)
(Y,-X,-Z)
(-Y,X,-Z)
(X,-Y,Z)
(-X,Y,Z)
(1/2-Y,1/2-X,1/2+Z)
(1/2+Y,1/2+X,1/2+Z)

P 42/N C M	16
(X,Y,Z)
(-X,-Y,Z)
(1/2-Y,1/2+X,1/2+Z)
(1/2+Y,1/2-X,1/2+Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,1/2-Y,-Z)
(Y,X,1/2-Z)
(-Y,-X,1/2-Z)
(1/2-X,1/2-Y,1/2-Z)
(1/2+X,1/2+Y,1/2-Z)
(Y,-X,-Z)
(-Y,X,-Z)
(X,-Y,1/2+Z)
(-X,Y,1/2+Z)
(1/2-Y,1/2-X,Z)
(1/2+Y,1/2+X,Z)

I 4/M M M	32
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,Z)
(Y,-X,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Y,X,-Z)
(-Y,-X,-Z)
(-X,-Y,-Z)
(X,Y,-Z)
(Y,-X,-Z)
(-Y,X,-Z)
(X,-Y,Z)
(-X,Y,Z)
(-Y,-X,Z)
(Y,X,Z)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,1/2-Y,1/2+Z)
(1/2-Y,1/2+X,1/2+Z)
(1/2+Y,1/2-X,1/2+Z)
(1/2-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2-Z)
(1/2+Y,1/2+X,1/2-Z)
(1/2-Y,1/2-X,1/2-Z)
(1/2-X,1/2-Y,1/2-Z)
(1/2+X,1/2+Y,1/2-Z)
(1/2+Y,1/2-X,1/2-Z)
(1/2-Y,1/2+X,1/2-Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2+Z)
(1/2-Y,1/2-X,1/2+Z)
(1/2+Y,1/2+X,1/2+Z)

I 4/M C M	32
(X,Y,Z)
(-X,-Y,Z)
(-Y,X,Z)
(Y,-X,Z)
(-X,Y,1/2-Z)
(X,-Y,1/2-Z)
(Y,X,1/2-Z)
(-Y,-X,1/2-Z)
(-X,-Y,-Z)
(X,Y,-Z)
(Y,-X,-Z)
(-Y,X,-Z)
(X,-Y,1/2+Z)
(-X,Y,1/2+Z)
(-Y,-X,1/2+Z)
(Y,X,1/2+Z)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,1/2-Y,1/2+Z)
(1/2-Y,1/2+X,1/2+Z)
(1/2+Y,1/2-X,1/2+Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,1/2-Y,-Z)
(1/2+Y,1/2+X,-Z)
(1/2-Y,1/2-X,-Z)
(1/2-X,1/2-Y,1/2-Z)
(1/2+X,1/2+Y,1/2-Z)
(1/2+Y,1/2-X,1/2-Z)
(1/2-Y,1/2+X,1/2-Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,Z)
(1/2-Y,1/2-X,Z)
(1/2+Y,1/2+X,Z)

I 41/A M D	32
(X,Y,Z)
(1/2-X,1/2-Y,1/2+Z)
(-Y,1/2+X,1/4+Z)
(1/2+Y,-X,3/4+Z)
(1/2-X,Y,3/4-Z)
(X,1/2-Y,1/4-Z)
(1/2+Y,1/2+X,1/2-Z)
(-Y,-X,-Z)
(-X,1/2-Y,1/4-Z)
(1/2+X,Y,3/4-Z)
(Y,-X,-Z)
(1/2-Y,1/2+X,1/2-Z)
(1/2+X,1/2-Y,1/2+Z)
(-X,Y,Z)
(1/2-Y,-X,3/4+Z)
(Y,1/2+X,1/4+Z)
(1/2+X,1/2+Y,1/2+Z)
(-X,-Y,Z)
(1/2-Y,X,3/4+Z)
(Y,1/2-X,1/4+Z)
(-X,1/2+Y,1/4-Z)
(1/2+X,-Y,3/4-Z)
(Y,X,-Z)
(1/2-Y,1/2-X,1/2-Z)
(1/2-X,-Y,3/4-Z)
(X,1/2+Y,1/4-Z)
(1/2+Y,1/2-X,1/2-Z)
(-Y,X,-Z)
(X,-Y,Z)
(1/2-X,1/2+Y,1/2+Z)
(-Y,1/2-X,1/4+Z)
(1/2+Y,X,3/4+Z)

I 41/A C D	32
(X,Y,Z)
(1/2-X,1/2-Y,1/2+Z)
(-Y,1/2+X,1/4+Z)
(1/2+Y,-X,3/4+Z)
(1/2-X,Y,1/4-Z)
(X,1/2-Y,3/4-Z)
(1/2+Y,1/2+X,-Z)
(-Y,-X,1/2-Z)
(-X,1/2-Y,1/4-Z)
(1/2+X,Y,3/4-Z)
(Y,-X,-Z)
(1/2-Y,1/2+X,1/2-Z)
(1/2+X,1/2-Y,Z)
(-X,Y,1/2+Z)
(1/2-Y,-X,1/4+Z)
(Y,1/2+X,3/4+Z)
(1/2+X,1/2+Y,1/2+Z)
(-X,-Y,Z)
(1/2-Y,X,3/4+Z)
(Y,1/2-X,1/4+Z)
(-X,1/2+Y,3/4-Z)
(1/2+X,-Y,1/4-Z)
(Y,X,1/2-Z)
(1/2-Y,1/2-X,-Z)
(1/2-X,-Y,3/4-Z)
(X,1/2+Y,1/4-Z)
(1/2+Y,1/2-X,1/2-Z)
(-Y,X,-Z)
(X,-Y,1/2+Z)
(1/2-X,1/2+Y,Z)
(-Y,1/2-X,3/4+Z)
(1/2+Y,X,1/4+Z)

P 3	3
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)

P 31	3
(X,Y,Z)
(-Y,X-Y,Z+1/3)
(Y-X,-X,Z+2/3)

P 32	3
(X,Y,Z)
(-Y,X-Y,Z+2/3)
(Y-X,-X,Z+1/3)

H 3	9
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(X+2/3,Y+1/3,Z+1/3)
(-Y+2/3,X-Y+1/3,Z+1/3)
(Y-X+2/3,-X+1/3,Z+1/3)
(X+1/3,Y+2/3,Z+2/3)
(-Y+1/3,X-Y+2/3,Z+2/3)
(Y-X+1/3,-X+2/3,Z+2/3)

P -3	6
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-X,-Y,-Z)
(Y,Y-X,-Z)
(X-Y,X,-Z)

R -3	18
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-X,-Y,-Z)
(Y,Y-X,-Z)
(X-Y,X,-Z)
(2/3+X,1/3+Y,1/3+Z)
(2/3-Y,1/3+X-Y,1/3+Z)
(2/3+Y-X,1/3-X,1/3+Z)
(2/3-X,1/3-Y,1/3-Z)
(2/3+Y,1/3+Y-X,1/3-Z)
(2/3+X-Y,1/3+X,1/3-Z)
(1/3+X,2/3+Y,2/3+Z)
(1/3-Y,2/3+X-Y,2/3+Z)
(1/3+Y-X,2/3-X,2/3+Z)
(1/3-X,2/3-Y,2/3-Z)
(1/3+Y,2/3+Y-X,2/3-Z)
(1/3+X-Y,2/3+X,2/3-Z)

P 3 1 2	6
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-Y,-X,-Z)
(Y-X,Y,-Z)
(X,X-Y,-Z)

P 3 2 1	6
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(Y,X,-Z)
(X-Y,-Y,-Z)
(-X,Y-X,-Z)

P 31 1 2	6
(X,Y,Z)
(-Y,X-Y,1/3+Z)
(Y-X,-X,2/3+Z)
(-Y,-X,2/3-Z)
(Y-X,Y,1/3-Z)
(X,X-Y,-Z)

P 31 2 1	6
(X,Y,Z)
(-Y,X-Y,Z+1/3)
(Y-X,-X,Z+2/3)
(Y,X,-Z)
(X-Y,-Y,2/3-Z)
(-X,Y-X,1/3-Z)

P 32 1 2	6
(X,Y,Z)
(-Y,X-Y,2/3+Z)
(Y-X,-X,1/3+Z)
(-Y,-X,1/3-Z)
(Y-X,Y,2/3-Z)
(X,X-Y,-Z)

P 32 2 1	6
(X,Y,Z)
(-Y,X-Y,Z+2/3)
(Y-X,-X,Z+1/3)
(Y,X,-Z)
(X-Y,-Y,1/3-Z)
(-X,Y-X,2/3-Z)

H 3 2	18
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(Y,X,-Z)
(X-Y,-Y,-Z)
(-X,Y-X,-Z)
(2/3+X,1/3+Y,1/3+Z)
(2/3-Y,1/3+X-Y,1/3+Z)
(2/3+Y-X,1/3-X,1/3+Z)
(2/3+Y,1/3+X,1/3-Z)
(2/3+X-Y,1/3-Y,1/3-Z)
(2/3-X,1/3+Y-X,1/3-Z)
(1/3+X,2/3+Y,2/3+Z)
(1/3-Y,2/3+X-Y,2/3+Z)
(1/3+Y-X,2/3-X,2/3+Z)
(1/3+Y,2/3+X,2/3-Z)
(1/3+X-Y,2/3-Y,2/3-Z)
(1/3-X,2/3+Y-X,2/3-Z)

P 3 M 1	6
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-Y,-X,Z)
(Y-X,Y,Z)
(X,X-Y,Z)

P 3 1 M	6
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(Y,X,Z)
(X-Y,-Y,Z)
(-X,Y-X,Z)

P 3 C 1	6
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-Y,-X,1/2+Z)
(Y-X,Y,1/2+Z)
(X,X-Y,1/2+Z)

P 3 1 C	6
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(Y,X,1/2+Z)
(X-Y,-Y,1/2+Z)
(-X,Y-X,1/2+Z)

R 3 M	18
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-Y,-X,Z)
(Y-X,Y,Z)
(X,X-Y,Z)
(2/3+X,1/3+Y,1/3+Z)
(2/3-Y,1/3+X-Y,1/3+Z)
(2/3+Y-X,1/3-X,1/3+Z)
(2/3-Y,1/3-X,1/3+Z)
(2/3+Y-X,1/3+Y,1/3+Z)
(2/3+X,1/3+X-Y,1/3+Z)
(1/3+X,2/3+Y,2/3+Z)
(1/3-Y,2/3+X-Y,2/3+Z)
(1/3+Y-X,2/3-X,2/3+Z)
(1/3-Y,2/3-X,2/3+Z)
(1/3+Y-X,2/3+Y,2/3+Z)
(1/3+X,2/3+X-Y,2/3+Z)

R 3 C	18
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-Y,-X,1/2+Z)
(Y-X,Y,1/2+Z)
(X,X-Y,1/2+Z)
(2/3+X,1/3+Y,1/3+Z)
(2/3-Y,1/3+X-Y,1/3+Z)
(2/3+Y-X,1/3-X,1/3+Z)
(2/3-Y,1/3-X,5/6+Z)
(2/3+Y-X,1/3+Y,5/6+Z)
(2/3+X,1/3+X-Y,5/6+Z)
(1/3+X,2/3+Y,2/3+Z)
(1/3-Y,2/3+X-Y,2/3+Z)
(1/3+Y-X,2/3-X,2/3+Z)
(1/3-Y,2/3-X,1/6+Z)
(1/3+Y-X,2/3+Y,1/6+Z)
(1/3+X,2/3+X-Y,1/6+Z)

P -3 1 M	12
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-Y,-X,-Z)
(Y-X,Y,-Z)
(X,X-Y,-Z)
(-X,-Y,-Z)
(Y,Y-X,-Z)
(X-Y,X,-Z)
(Y,X,Z)
(X-Y,-Y,Z)
(-X,Y-X,Z)

P -3 1 C	12
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-Y,-X,1/2-Z)
(Y-X,Y,1/2-Z)
(X,X-Y,1/2-Z)
(-X,-Y,-Z)
(Y,Y-X,-Z)
(X-Y,X,-Z)
(Y,X,1/2+Z)
(X-Y,-Y,1/2+Z)
(-X,Y-X,1/2+Z)

P -3 M 1	12
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(Y,X,-Z)
(X-Y,-Y,-Z)
(-X,Y-X,-Z)
(-X,-Y,-Z)
(Y,Y-X,-Z)
(X-Y,X,-Z)
(-Y,-X,Z)
(Y-X,Y,Z)
(X,X-Y,Z)

P -3 C 1	12
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(Y,X,1/2-Z)
(X-Y,-Y,1/2-Z)
(-X,Y-X,1/2-Z)
(-X,-Y,-Z)
(Y,Y-X,-Z)
(X-Y,X,-Z)
(-Y,-X,1/2+Z)
(Y-X,Y,1/2+Z)
(X,X-Y,1/2+Z)

R -3 M	36
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(Y,X,-Z)
(X-Y,-Y,-Z)
(-X,Y-X,-Z)
(-X,-Y,-Z)
(Y,Y-X,-Z)
(X-Y,X,-Z)
(-Y,-X,Z)
(Y-X,Y,Z)
(X,X-Y,Z)
(2/3+X,1/3+Y,1/3+Z)
(2/3-Y,1/3+X-Y,1/3+Z)
(2/3+Y-X,1/3-X,1/3+Z)
(2/3+Y,1/3+X,1/3-Z)
(2/3+X-Y,1/3-Y,1/3-Z)
(2/3-X,1/3+Y-X,1/3-Z)
(2/3-X,1/3-Y,1/3-Z)
(2/3+Y,1/3+Y-X,1/3-Z)
(2/3+X-Y,1/3+X,1/3-Z)
(2/3-Y,1/3-X,1/3+Z)
(2/3+Y-X,1/3+Y,1/3+Z)
(2/3+X,1/3+X-Y,1/3+Z)
(1/3+X,2/3+Y,2/3+Z)
(1/3-Y,2/3+X-Y,2/3+Z)
(1/3+Y-X,2/3-X,2/3+Z)
(1/3+Y,2/3+X,2/3-Z)
(1/3+X-Y,2/3-Y,2/3-Z)
(1/3-X,2/3+Y-X,2/3-Z)
(1/3-X,2/3-Y,2/3-Z)
(1/3+Y,2/3+Y-X,2/3-Z)
(1/3X-Y,2/3+X,2/3-Z)
(1/3-Y,2/3-X,2/3+Z)
(1/3+Y-X,2/3+Y,2/3+Z)
(1/3+X,2/3+X-Y,2/3+Z)

R -3 C	36
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(Y,X,1/2-Z)
(X-Y,-Y,1/2-Z)
(-X,Y-X,1/2-Z)
(-X,-Y,-Z)
(Y,Y-X,-Z)
(X-Y,X,-Z)
(-Y,-X,1/2+Z)
(Y-X,Y,1/2+Z)
(X,X-Y,1/2+Z)
(2/3+X,1/3+Y,1/3+Z)
(2/3-Y,1/3+X-Y,1/3+Z)
(2/3+Y-X,1/3-X,1/3+Z)
(2/3+Y,1/3+X,5/6-Z)
(2/3+X-Y,1/3-Y,5/6-Z)
(2/3-X,1/3+Y-X,5/6-Z)
(2/3-X,1/3-Y,1/3-Z)
(2/3+Y,1/3+Y-X,1/3-Z)
(2/3+X-Y,1/3+X,1/3-Z)
(2/3-Y,1/3-X,5/6+Z)
(2/3+Y-X,1/3+Y,5/6+Z)
(2/3+X,1/3+X-Y,5/6+Z)
(1/3+X,2/3+Y,2/3+Z)
(1/3-Y,2/3+X-Y,2/3+Z)
(1/3+Y-X,2/3-X,2/3+Z)
(1/3+Y,2/3+X,1/6-Z)
(1/3+X-Y,2/3-Y,1/6-Z)
(1/3-X,2/3+Y-X,1/6-Z)
(1/3-X,2/3-Y,2/3-Z)
(1/3+Y,2/3+Y-X,2/3-Z)
(1/3+X-Y,2/3+X,2/3-Z)
(1/3-Y,2/3-X,1/6+Z)
(1/3+Y-X,2/3+Y,1/6+Z)
(1/3+X,2/3+X-Y,1/6+Z)

P 6	6
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-X,-Y,Z)
(Y,Y-X,Z)
(X-Y,X,Z)

P 61	6
(X,Y,Z)
(-Y,X-Y,Z+1/3)
(Y-X,-X,Z+2/3)
(-X,-Y,Z+1/2)
(Y,Y-X,Z+5/6)
(X-Y,X,Z+1/6)

P 65	6
(X,Y,Z)
(-Y,X-Y,Z+2/3)
(Y-X,-X,Z+1/3)
(-X,-Y,Z+1/2)
(Y,Y-X,Z+1/6)
(X-Y,X,Z+5/6)

P 62	6
(X,Y,Z)
(-Y,X-Y,2/3+Z)
(Y-X,-X,1/3+Z)
(-X,-Y,Z)
(Y,Y-X,2/3+Z)
(X-Y,X,1/3+Z)

P 64	6
(X,Y,Z)
(-Y,X-Y,1/3+Z)
(Y-X,-X,2/3+Z)
(-X,-Y,Z)
(Y,Y-X,1/3+Z)
(X-Y,X,2/3+Z)

P 63	6
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-X,-Y,1/2+Z)
(Y,Y-X,1/2+Z)
(X-Y,X,1/2+Z)

P -6	6
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(X,Y,-Z)
(-Y,X-Y,-Z)
(Y-X,-X,-Z)

P 6/M	12
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-X,-Y,Z)
(Y,Y-X,Z)
(X-Y,X,Z)
(-X,-Y,-Z)
(Y,Y-X,-Z)
(X-Y,X,-Z)
(X,Y,-Z)
(-Y,X-Y,-Z)
(Y-X,-X,-Z)

P 63/M	12
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-X,-Y,1/2+Z)
(Y,Y-X,1/2+Z)
(X-Y,X,1/2+Z)
(-X,-Y,-Z)
(Y,Y-X,-Z)
(X-Y,X,-Z)
(X,Y,1/2-Z)
(-Y,X-Y,1/2-Z)
(Y-X,-X,1/2-Z)

P 6 2 2	12
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-X,-Y,Z)
(Y,Y-X,Z)
(X-Y,X,Z)
(Y,X,-Z)
(X-Y,-Y,-Z)
(-X,Y-X,-Z)
(-Y,-X,-Z)
(Y-X,Y,-Z)
(X,X-Y,-Z)

P 61 2 2	12
(X,Y,Z)
(-Y,X-Y,1/3+Z)
(Y-X,-X,2/3+Z)
(-X,-Y,1/2+Z)
(Y,Y-X,5/6+Z)
(X-Y,X,1/6+Z)
(Y,X,1/3-Z)
(X-Y,-Y,-Z)
(-X,Y-X,2/3-Z)
(-Y,-X,5/6-Z)
(Y-X,Y,1/2-Z)
(X,X-Y,1/6-Z)

P 65 2 2	12
(X,Y,Z)
(-Y,X-Y,2/3+Z)
(Y-X,-X,1/3+Z)
(-X,-Y,1/2+Z)
(Y,Y-X,1/6+Z)
(X-Y,X,5/6+Z)
(Y,X,2/3-Z)
(X-Y,-Y,-Z)
(-X,Y-X,1/3-Z)
(-Y,-X,1/6-Z)
(Y-X,Y,1/2-Z)
(X,X-Y,5/6-Z)

P 62 2 2	12
(X,Y,Z)
(-Y,X-Y,2/3+Z)
(Y-X,-X,1/3+Z)
(-X,-Y,Z)
(Y,Y-X,2/3+Z)
(X-Y,X,1/3+Z)
(Y,X,2/3-Z)
(X-Y,-Y,-Z)
(-X,Y-X,1/3-Z)
(-Y,-X,2/3-Z)
(Y-X,Y,-Z)
(X,X-Y,1/3-Z)

P 64 2 2	12
(X,Y,Z)
(-Y,X-Y,1/3+Z)
(Y-X,-X,2/3+Z)
(-X,-Y,Z)
(Y,Y-X,1/3+Z)
(X-Y,X,2/3+Z)
(Y,X,1/3-Z)
(X-Y,-Y,-Z)
(-X,Y-X,2/3-Z)
(-Y,-X,1/3-Z)
(Y-X,Y,-Z)
(X,X-Y,2/3-Z)

P 63 2 2	12
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-X,-Y,1/2+Z)
(Y,Y-X,1/2+Z)
(X-Y,X,1/2+Z)
(Y,X,-Z)
(X-Y,-Y,-Z)
(-X,Y-X,-Z)
(-Y,-X,1/2-Z)
(Y-X,Y,1/2-Z)
(X,X-Y,1/2-Z)

P 6 M M	12
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-X,-Y,Z)
(Y,Y-X,Z)
(X-Y,X,Z)
(-Y,-X,Z)
(Y-X,Y,Z)
(X,X-Y,Z)
(Y,X,Z)
(X-Y,-Y,Z)
(-X,Y-X,Z)

P 6 C C	12
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-X,-Y,Z)
(Y,Y-X,Z)
(X-Y,X,Z)
(-Y,-X,1/2+Z)
(Y-X,Y,1/2+Z)
(X,X-Y,1/2+Z)
(Y,X,1/2+Z)
(X-Y,-Y,1/2+Z)
(-X,Y-X,1/2+Z)

P 63 C M	12
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-X,-Y,1/2+Z)
(Y,Y-X,1/2+Z)
(X-Y,X,1/2+Z)
(-Y,-X,1/2+Z)
(Y-X,Y,1/2+Z)
(X,X-Y,1/2+Z)
(Y,X,Z)
(X-Y,-Y,Z)
(-X,Y-X,Z)

P 63 M C	12
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-X,-Y,1/2+Z)
(Y,Y-X,1/2+Z)
(X-Y,X,1/2+Z)
(-Y,-X,Z)
(Y-X,Y,Z)
(X,X-Y,Z)
(Y,X,1/2+Z)
(X-Y,-Y,1/2+Z)
(-X,Y-X,1/2+Z)

P -6 M 2	12
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(X,Y,-Z)
(-Y,X-Y,-Z)
(Y-X,-X,-Z)
(-Y,-X,Z)
(Y-X,Y,Z)
(X,X-Y,Z)
(-Y,-X,-Z)
(Y-X,Y,-Z)
(X,X-Y,-Z)

P -6 C 2	12
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(X,Y,1/2-Z)
(-Y,X-Y,1/2-Z)
(Y-X,-X,1/2-Z)
(-Y,-X,1/2+Z)
(Y-X,Y,1/2+Z)
(X,X-Y,1/2+Z)
(-Y,-X,-Z)
(Y-X,Y,-Z)
(X,X-Y,-Z)

P -6 2 M	12
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(X,Y,-Z)
(-Y,X-Y,-Z)
(Y-X,-X,-Z)
(Y,X,-Z)
(X-Y,-Y,-Z)
(-X,Y-X,-Z)
(Y,X,Z)
(X-Y,-Y,Z)
(-X,Y-X,Z)

P -6 2 C	12
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(X,Y,1/2-Z)
(-Y,X-Y,1/2-Z)
(Y-X,-X,1/2-Z)
(Y,X,-Z)
(X-Y,-Y,-Z)
(-X,Y-X,-Z)
(Y,X,1/2+Z)
(X-Y,-Y,1/2+Z)
(-X,Y-X,1/2+Z)

P 6/M M M	24
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-X,-Y,Z)
(Y,Y-X,Z)
(X-Y,X,Z)
(Y,X,-Z)
(X-Y,-Y,-Z)
(-X,Y-X,-Z)
(-Y,-X,-Z)
(Y-X,Y,-Z)
(X,X-Y,-Z)
(-X,-Y,-Z)
(Y,Y-X,-Z)
(X-Y,X,-Z)
(X,Y,-Z)
(Y-X,-X,-Z)
(-Y,X-Y,-Z)
(-Y,-X,Z)
(Y-X,Y,Z)
(X,X-Y,Z)
(Y,X,Z)
(X-Y,-Y,Z)
(-X,Y-X,Z)

P 6/M C C	24
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-X,-Y,Z)
(Y,Y-X,Z)
(X-Y,X,Z)
(Y,X,1/2-Z)
(X-Y,-Y,1/2-Z)
(-X,Y-X,1/2-Z)
(-Y,-X,1/2-Z)
(Y-X,Y,1/2-Z)
(X,X-Y,1/2-Z)
(-X,-Y,-Z)
(Y,Y-X,-Z)
(X-Y,X,-Z)
(X,Y,-Z)
(Y-X,-X,-Z)
(-Y,X-Y,-Z)
(-Y,-X,1/2+Z)
(Y-X,Y,1/2+Z)
(X,X-Y,1/2+Z)
(Y,X,1/2+Z)
(X-Y,-Y,1/2+Z)
(-X,Y-X,1/2+Z)

P 63/M C M	24
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-X,-Y,1/2+Z)
(Y,Y-X,1/2+Z)
(X-Y,X,1/2+Z)
(Y,X,1/2-Z)
(X-Y,-Y,1/2-Z)
(-X,Y-X,1/2-Z)
(-Y,-X,-Z)
(Y-X,Y,-Z)
(X,X-Y,-Z)
(-X,-Y,-Z)
(Y,Y-X,-Z)
(X-Y,X,-Z)
(X,Y,1/2-Z)
(Y-X,-X,1/2-Z)
(-Y,X-Y,1/2-Z)
(-Y,-X,1/2+Z)
(Y-X,Y,1/2+Z)
(X,X-Y,1/2+Z)
(Y,X,Z)
(X-Y,-Y,Z)
(-X,Y-X,Z)

P 63/M M C	24
(X,Y,Z)
(-Y,X-Y,Z)
(Y-X,-X,Z)
(-X,-Y,1/2+Z)
(Y,Y-X,1/2+Z)
(X-Y,X,1/2+Z)
(Y,X,-Z)
(X-Y,-Y,-Z)
(-X,Y-X,-Z)
(-Y,-X,1/2-Z)
(Y-X,Y,1/2-Z)
(X,X-Y,1/2-Z)
(-X,-Y,-Z)
(Y,Y-X,-Z)
(X-Y,X,-Z)
(X,Y,1/2-Z)
(Y-X,-X,1/2-Z)
(-Y,X-Y,1/2-Z)
(-Y,-X,Z)
(Y-X,Y,Z)
(X,X-Y,Z)
(Y,X,1/2+Z)
(X-Y,-Y,1/2+Z)
(-X,Y-X,1/2+Z)

P 2 3	12
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)

F 2 3	48
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(X,1/2+Y,1/2+Z)
(-X,1/2-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(X,1/2-Y,1/2-Z)
(Z,1/2+X,1/2+Y)
(Z,1/2-X,1/2-Y)
(-Z,1/2-X,1/2+Y)
(-Z,1/2+X,1/2-Y)
(Y,1/2+Z,1/2+X)
(-Y,1/2+Z,1/2-X)
(Y,1/2-Z,1/2-X)
(-Y,1/2-Z,1/2+X)
(1/2+X,Y,1/2+Z)
(1/2-X,-Y,1/2+Z)
(1/2-X,Y,1/2-Z)
(1/2+X,-Y,1/2-Z)
(1/2+Z,X,1/2+Y)
(1/2+Z,-X,1/2-Y)
(1/2-Z,-X,1/2+Y)
(1/2-Z,X,1/2-Y)
(1/2+Y,Z,1/2+X)
(1/2-Y,Z,1/2-X)
(1/2+Y,-Z,1/2-X)
(1/2-Y,-Z,1/2+X)
(1/2+X,1/2+Y,Z)
(1/2-X,1/2-Y,Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,1/2-Y,-Z)
(1/2+Z,1/2+X,Y)
(1/2+Z,1/2-X,-Y)
(1/2-Z,1/2-X,Y)
(1/2-Z,1/2+X,-Y)
(1/2+Y,1/2+Z,X)
(1/2-Y,1/2+Z,-X)
(1/2+Y,1/2-Z,-X)
(1/2-Y,1/2-Z,X)

I 2 3	24
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2-Z)
(1/2+Z,1/2+X,1/2+Y)
(1/2+Z,1/2-X,1/2-Y)
(1/2-Z,1/2-X,1/2+Y)
(1/2-Z,1/2+X,1/2-Y)
(1/2+Y,1/2+Z,1/2+X)
(1/2-Y,1/2+Z,1/2-X)
(1/2+Y,1/2-Z,1/2-X)
(1/2-Y,1/2-Z,1/2+X)

P 21 3	12
(X,Y,Z)
(1/2-X,-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,-Z)
(Z,X,Y)
(1/2+Z,1/2-X,-Y)
(1/2-Z,-X,1/2+Y)
(-Z,1/2+X,1/2-Y)
(Y,Z,X)
(-Y,1/2+Z,1/2-X)
(1/2+Y,1/2-Z,-X)
(1/2-Y,-Z,1/2+X)

I 21 3	24
(X,Y,Z)
(1/2-X,-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,-Z)
(Z,X,Y)
(1/2+Z,1/2-X,-Y)
(1/2-Z,-X,1/2+Y)
(-Z,1/2+X,1/2-Y)
(Y,Z,X)
(-Y,1/2+Z,1/2-X)
(1/2+Y,1/2-Z,-X)
(1/2-Y,-Z,1/2+X)
(1/2+X,1/2+Y,1/2+Z)
(-X,1/2-Y,Z)
(1/2-X,Y,-Z)
(X,-Y,1/2-Z)
(1/2+Z,1/2+X,1/2+Y)
(Z,-X,1/2-Y)
(-Z,1/2-X,Y)
(1/2-Z,X,-Y)
(1/2+Y,1/2+Z,1/2+X)
(1/2-Y,Z,-X)
(Y,-Z,1/2-X)
(-Y,1/2-Z,X)

P M -3	24
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(-X,-Y,-Z)
(X,Y,-Z)
(X,-Y,Z)
(-X,Y,Z)
(-Z,-X,-Y)
(-Z,X,Y)
(Z,X,-Y)
(Z,-X,Y)
(-Y,-Z,-X)
(Y,-Z,X)
(-Y,Z,X)
(Y,Z,-X)

P N -3	24
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(1/2-X,1/2-Y,1/2-Z)
(1/2+X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2+Z)
(1/2-Z,1/2-X,1/2-Y)
(1/2-Z,1/2+X,1/2+Y)
(1/2+Z,1/2+X,1/2-Y)
(1/2+Z,1/2-X,1/2+Y)
(1/2-Y,1/2-Z,1/2-X)
(1/2+Y,1/2-Z,1/2+X)
(1/2-Y,1/2+Z,1/2+X)
(1/2+Y,1/2+Z,1/2-X)

F M -3	96
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(-X,-Y,-Z)
(X,Y,-Z)
(X,-Y,Z)
(-X,Y,Z)
(-Z,-X,-Y)
(-Z,X,Y)
(Z,X,-Y)
(Z,-X,Y)
(-Y,-Z,-X)
(Y,-Z,X)
(-Y,Z,X)
(Y,Z,-X)
(X,1/2+Y,1/2+Z)
(-X,1/2-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(X,1/2-Y,1/2-Z)
(Z,1/2+X,1/2+Y)
(Z,1/2-X,1/2-Y)
(-Z,1/2-X,1/2+Y)
(-Z,1/2+X,1/2-Y)
(Y,1/2+Z,1/2+X)
(-Y,1/2+Z,1/2-X)
(Y,1/2-Z,1/2-X)
(-Y,1/2-Z,1/2+X)
(-X,1/2-Y,1/2-Z)
(X,1/2+Y,1/2-Z)
(X,1/2-Y,1/2+Z)
(-X,1/2+Y,1/2+Z)
(-Z,1/2-X,1/2-Y)
(-Z,1/2+X,1/2+Y)
(Z,1/2+X,1/2-Y)
(Z,1/2-X,1/2+Y)
(-Y,1/2-Z,1/2-X)
(Y,1/2-Z,1/2+X)
(-Y,1/2+Z,1/2+X)
(Y,1/2+Z,1/2-X)
(1/2+X,Y,1/2+Z)
(1/2-X,-Y,1/2+Z)
(1/2-X,Y,1/2-Z)
(1/2+X,-Y,1/2-Z)
(1/2+Z,X,1/2+Y)
(1/2+Z,-X,1/2-Y)
(1/2-Z,-X,1/2+Y)
(1/2-Z,X,1/2-Y)
(1/2+Y,Z,1/2+X)
(1/2-Y,Z,1/2-X)
(1/2+Y,-Z,1/2-X)
(1/2-Y,-Z,1/2+X)
(1/2-X,-Y,1/2-Z)
(1/2+X,Y,1/2-Z)
(1/2+X,-Y,1/2+Z)
(1/2-X,Y,1/2+Z)
(1/2-Z,-X,1/2-Y)
(1/2-Z,X,1/2+Y)
(1/2+Z,X,1/2-Y)
(1/2+Z,-X,1/2+Y)
(1/2-Y,-Z,1/2-X)
(1/2+Y,-Z,1/2+X)
(1/2-Y,Z,1/2+X)
(1/2+Y,Z,1/2-X)
(1/2+X,1/2+Y,Z)
(1/2-X,1/2-Y,Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,1/2-Y,-Z)
(1/2+Z,1/2+X,Y)
(1/2+Z,1/2-X,-Y)
(1/2-Z,1/2-X,Y)
(1/2-Z,1/2+X,-Y)
(1/2+Y,1/2+Z,X)
(1/2-Y,1/2+Z,-X)
(1/2+Y,1/2-Z,-X)
(1/2-Y,1/2-Z,X)
(1/2-X,1/2-Y,-Z)
(1/2+X,1/2+Y,-Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,Z)
(1/2-Z,1/2-X,-Y)
(1/2-Z,1/2+X,Y)
(1/2+Z,1/2+X,-Y)
(1/2+Z,1/2-X,Y)
(1/2-Y,1/2-Z,-X)
(1/2+Y,1/2-Z,X)
(1/2-Y,1/2+Z,X)
(1/2+Y,1/2+Z,-X)

F D -3	96
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(1/4-X,1/4-Y,1/4-Z)
(1/4+X,1/4+Y,1/4-Z)
(1/4+X,1/4-Y,1/4+Z)
(1/4-X,1/4+Y,1/4+Z)
(1/4-Z,1/4-X,1/4-Y)
(1/4-Z,1/4+X,1/4+Y)
(1/4+Z,1/4+X,1/4-Y)
(1/4+Z,1/4-X,1/4+Y)
(1/4-Y,1/4-Z,1/4-X)
(1/4+Y,1/4-Z,1/4+X)
(1/4-Y,1/4+Z,1/4+X)
(1/4+Y,1/4+Z,1/4-X)
(X,1/2+Y,1/2+Z)
(-X,1/2-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(X,1/2-Y,1/2-Z)
(Z,1/2+X,1/2+Y)
(Z,1/2-X,1/2-Y)
(-Z,1/2-X,1/2+Y)
(-Z,1/2+X,1/2-Y)
(Y,1/2+Z,1/2+X)
(-Y,1/2+Z,1/2-X)
(Y,1/2-Z,1/2-X)
(-Y,1/2-Z,1/2+X)
(1/4-X,3/4-Y,3/4-Z)
(1/4+X,3/4+Y,3/4-Z)
(1/4+X,3/4-Y,3/4+Z)
(1/4-X,3/4+Y,3/4+Z)
(1/4-Z,3/4-X,3/4-Y)
(1/4-Z,3/4+X,3/4+Y)
(1/4+Z,3/4+X,3/4-Y)
(1/4+Z,3/4-X,3/4+Y)
(1/4-Y,3/4-Z,3/4-X)
(1/4+Y,3/4-Z,3/4+X)
(1/4-Y,3/4+Z,3/4+X)
(1/4+Y,3/4+Z,3/4-X)
(1/2+X,Y,1/2+Z)
(1/2-X,-Y,1/2+Z)
(1/2-X,Y,1/2-Z)
(1/2+X,-Y,1/2-Z)
(1/2+Z,X,1/2+Y)
(1/2+Z,-X,1/2-Y)
(1/2-Z,-X,1/2+Y)
(1/2-Z,X,1/2-Y)
(1/2+Y,Z,1/2+X)
(1/2-Y,Z,1/2-X)
(1/2+Y,-Z,1/2-X)
(1/2-Y,-Z,1/2+X)
(3/4-X,1/4-Y,3/4-Z)
(3/4+X,1/4+Y,3/4-Z)
(3/4+X,1/4-Y,3/4+Z)
(3/4-X,1/4+Y,3/4+Z)
(3/4-Z,1/4-X,3/4-Y)
(3/4-Z,1/4+X,3/4+Y)
(3/4+Z,1/4+X,3/4-Y)
(3/4+Z,1/4-X,3/4+Y)
(3/4-Y,1/4-Z,3/4-X)
(3/4+Y,1/4-Z,3/4+X)
(3/4-Y,1/4+Z,3/4+X)
(3/4+Y,1/4+Z,3/4-X)
(1/2+X,1/2+Y,Z)
(1/2-X,1/2-Y,Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,1/2-Y,-Z)
(1/2+Z,1/2+X,Y)
(1/2+Z,1/2-X,-Y)
(1/2-Z,1/2-X,Y)
(1/2-Z,1/2+X,-Y)
(1/2+Y,1/2+Z,X)
(1/2-Y,1/2+Z,-X)
(1/2+Y,1/2-Z,-X)
(1/2-Y,1/2-Z,X)
(3/4-X,3/4-Y,1/4-Z)
(3/4+X,3/4+Y,1/4-Z)
(3/4+X,3/4-Y,Z+1/4)
(3/4-X,3/4+Y,Z+1/4)
(3/4-Z,3/4-X,1/4-Y)
(3/4-Z,3/4+X,1/4+Y)
(3/4+Z,3/4+X,1/4-Y)
(3/4+Z,3/4-X,1/4+Y)
(3/4-Y,3/4-Z,1/4-X)
(3/4+Y,3/4-Z,1/4+X)
(3/4-Y,3/4+Z,1/4+X)
(3/4+Y,3/4+Z,1/4-X)

I M -3	48
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(-X,-Y,-Z)
(X,Y,-Z)
(X,-Y,Z)
(-X,Y,Z)
(-Z,-X,-Y)
(-Z,X,Y)
(Z,X,-Y)
(Z,-X,Y)
(-Y,-Z,-X)
(Y,-Z,X)
(-Y,Z,X)
(Y,Z,-X)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2-Z)
(1/2+Z,1/2+X,1/2+Y)
(1/2+Z,1/2-X,1/2-Y)
(1/2-Z,1/2-X,1/2+Y)
(1/2-Z,1/2+X,1/2-Y)
(1/2+Y,1/2+Z,1/2+X)
(1/2-Y,1/2+Z,1/2-X)
(1/2+Y,1/2-Z,1/2-X)
(1/2-Y,1/2-Z,1/2+X)
(1/2-X,1/2-Y,1/2-Z)
(1/2+X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2+Z)
(1/2-Z,1/2-X,1/2-Y)
(1/2-Z,1/2+X,1/2+Y)
(1/2+Z,1/2+X,1/2-Y)
(1/2+Z,1/2-X,1/2+Y)
(1/2-Y,1/2-Z,1/2-X)
(1/2+Y,1/2-Z,1/2+X)
(1/2-Y,1/2+Z,1/2+X)
(1/2+Y,1/2+Z,1/2-X)

P A -3	24
(X,Y,Z)
(1/2-X,-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,-Z)
(Z,X,Y)
(1/2+Z,1/2-X,-Y)
(1/2-Z,-X,1/2+Y)
(-Z,1/2+X,1/2-Y)
(Y,Z,X)
(-Y,1/2+Z,1/2-X)
(1/2+Y,1/2-Z,-X)
(1/2-Y,-Z,1/2+X)
(-X,-Y,-Z)
(1/2+X,Y,1/2-Z)
(X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,Z)
(-Z,-X,-Y)
(1/2-Z,1/2+X,Y)
(1/2+Z,X,1/2-Y)
(Z,1/2-X,1/2+Y)
(-Y,-Z,-X)
(Y,1/2-Z,1/2+X)
(1/2-Y,1/2+Z,X)
(1/2+Y,Z,1/2-X)

I A -3	48
(X,Y,Z)
(1/2-X,-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,-Z)
(Z,X,Y)
(1/2+Z,1/2-X,-Y)
(1/2-Z,-X,1/2+Y)
(-Z,1/2+X,1/2-Y)
(Y,Z,X)
(-Y,1/2+Z,1/2-X)
(1/2+Y,1/2-Z,-X)
(1/2-Y,-Z,1/2+X)
(-X,-Y,-Z)
(1/2+X,Y,1/2-Z)
(X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,Z)
(-Z,-X,-Y)
(1/2-Z,1/2+X,Y)
(1/2+Z,X,1/2-Y)
(Z,1/2-X,1/2+Y)
(-Y,-Z,-X)
(Y,1/2-Z,1/2+X)
(1/2-Y,1/2+Z,X)
(1/2+Y,Z,1/2-X)
(1/2+X,1/2+Y,1/2+Z)
(-X,1/2-Y,Z)
(1/2-X,+Y,-Z)
(X,-Y,1/2-Z)
(1/2+Z,1/2+X,1/2+Y)
(Z,-X,1/2-Y)
(-Z,1/2-X,Y)
(1/2-Z,X,-Y)
(1/2+Y,1/2+Z,1/2+X)
(1/2-Y,Z,-X)
(Y,-Z,1/2-X)
(-Y,1/2-Z,X)
(1/2-X,1/2-Y,1/2-Z)
(X,1/2+Y,-Z)
(1/2+X,-Y,Z)
(-X,Y,1/2+Z)
(1/2-Z,1/2-X,1/2-Y)
(-Z,X,1/2+Y)
(Z,1/2+X,-Y)
(1/2+Z,-X,Y)
(1/2-Y,1/2-Z,1/2-X)
(1/2+Y,-Z,X)
(-Y,Z,1/2+X)
(Y,1/2+Z,-X)

P 4 3 2	24
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(Y,X,-Z)
(-Y,-X,-Z)
(Y,-X,Z)
(-Y,X,Z)
(X,Z,-Y)
(-X,Z,Y)
(-X,-Z,-Y)
(X,-Z,Y)
(Z,Y,-X)
(Z,-Y,X)
(-Z,Y,X)
(-Z,-Y,-X)

P 42 3 2	24
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(1/2+Y,1/2+X,1/2-Z)
(1/2-Y,1/2-X,1/2-Z)
(1/2+Y,1/2-X,1/2+Z)
(1/2-Y,1/2+X,1/2+Z)
(1/2+X,1/2+Z,1/2-Y)
(1/2-X,1/2+Z,1/2+Y)
(1/2-X,1/2-Z,1/2-Y)
(1/2+X,1/2-Z,1/2+Y)
(1/2+Z,1/2+Y,1/2-X)
(1/2+Z,1/2-Y,1/2+X)
(1/2-Z,1/2+Y,1/2+X)
(1/2-Z,1/2-Y,1/2-X)

F 4 3 2	96
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(Y,X,-Z)
(-Y,-X,-Z)
(Y,-X,Z)
(-Y,X,Z)
(X,Z,-Y)
(-X,Z,Y)
(-X,-Z,-Y)
(X,-Z,Y)
(Z,Y,-X)
(Z,-Y,X)
(-Z,Y,X)
(-Z,-Y,-X)
(X,1/2+Y,1/2+Z)
(-X,1/2-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(X,1/2-Y,1/2-Z)
(Z,1/2+X,1/2+Y)
(Z,1/2-X,1/2-Y)
(-Z,1/2-X,1/2+Y)
(-Z,1/2+X,1/2-Y)
(Y,1/2+Z,1/2+X)
(-Y,1/2+Z,1/2-X)
(Y,1/2-Z,1/2-X)
(-Y,1/2-Z,1/2+X)
(Y,1/2+X,1/2-Z)
(-Y,1/2-X,1/2-Z)
(Y,1/2-X,1/2+Z)
(-Y,1/2+X,1/2+Z)
(X,1/2+Z,1/2-Y)
(-X,1/2+Z,1/2+Y)
(-X,1/2-Z,1/2-Y)
(X,1/2-Z,1/2+Y)
(Z,1/2+Y,1/2-X)
(Z,1/2-Y,1/2+X)
(-Z,1/2+Y,1/2+X)
(-Z,1/2-Y,1/2-X)
(1/2+X,Y,1/2+Z)
(1/2-X,-Y,1/2+Z)
(1/2-X,Y,1/2-Z)
(1/2+X,-Y,1/2-Z)
(1/2+Z,X,1/2+Y)
(1/2+Z,-X,1/2-Y)
(1/2-Z,-X,1/2+Y)
(1/2-Z,X,1/2-Y)
(1/2+Y,Z,1/2+X)
(1/2-Y,Z,1/2-X)
(1/2+Y,-Z,1/2-X)
(1/2-Y,-Z,1/2+X)
(1/2+Y,X,1/2-Z)
(1/2-Y,-X,1/2-Z)
(1/2+Y,-X,1/2+Z)
(1/2-Y,X,1/2+Z)
(1/2+X,Z,1/2-Y)
(1/2-X,Z,1/2+Y)
(1/2-X,-Z,1/2-Y)
(1/2+X,-Z,1/2+Y)
(1/2+Z,Y,1/2-X)
(1/2+Z,-Y,1/2+X)
(1/2-Z,Y,1/2+X)
(1/2-Z,-Y,1/2-X)
(1/2+X,1/2+Y,Z)
(1/2-X,1/2-Y,Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,1/2-Y,-Z)
(1/2+Z,1/2+X,Y)
(1/2+Z,1/2-X,-Y)
(1/2-Z,1/2-X,Y)
(1/2-Z,1/2+X,-Y)
(1/2+Y,1/2+Z,X)
(1/2-Y,1/2+Z,-X)
(1/2+Y,1/2-Z,-X)
(1/2-Y,1/2-Z,X)
(1/2+Y,1/2+X,-Z)
(1/2-Y,1/2-X,-Z)
(1/2+Y,1/2-X,Z)
(1/2-Y,1/2+X,Z)
(1/2+X,1/2+Z,-Y)
(1/2-X,1/2+Z,Y)
(1/2-X,1/2-Z,-Y)
(1/2+X,1/2-Z,Y)
(1/2+Z,1/2+Y,-X)
(1/2+Z,1/2-Y,X)
(1/2-Z,1/2+Y,X)
(1/2-Z,1/2-Y,-X)

F 41 3 2	96
(X,Y,Z)
(-X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,-Y,1/2-Z)
(Z,X,Y)
(1/2+Z,-X,1/2-Y)
(-Z,1/2-X,1/2+Y)
(1/2-Z,1/2+X,-Y)
(Y,Z,X)
(1/2-Y,1/2+Z,-X)
(1/2+Y,-Z,1/2-X)
(-Y,1/2-Z,1/2+X)
(3/4+Y,1/4+X,3/4-Z)
(1/4-Y,1/4-X,1/4-Z)
(1/4+Y,3/4-X,3/4+Z)
(3/4-Y,3/4+X,1/4+Z)
(3/4+X,1/4+Z,3/4-Y)
(3/4-X,3/4+Z,1/4+Y)
(1/4-X,1/4-Z,1/4-Y)
(1/4+X,3/4-Z,3/4+Y)
(3/4+Z,1/4+Y,3/4-X)
(1/4+Z,3/4-Y,3/4+X)
(3/4-Z,3/4+Y,1/4+X)
(1/4-Z,1/4-Y,1/4-X)
(X,1/2+Y,1/2+Z)
(-X,-Y,Z)
(1/2-X,Y,1/2-Z)
(1/2+X,1/2-Y,-Z)
(Z,1/2+X,1/2+Y)
(1/2+Z,1/2-X,-Y)
(-Z,-X,Y)
(1/2-Z,X,1/2-Y)
(Y,1/2+Z,1/2+X)
(1/2-Y,Z,1/2-X)
(1/2+Y,1/2-Z,-X)
(-Y,-Z,X)
(3/4+Y,3/4+X,1/4-Z)
(1/4-Y,3/4-X,3/4-Z)
(1/4+Y,1/4-X,1/4+Z)
(3/4-Y,1/4+X,3/4+Z)
(3/4+X,3/4+Z,1/4-Y)
(3/4-X,1/4+Z,3/4+Y)
(1/4-X,3/4-Z,3/4-Y)
(1/4+X,1/4-Z,1/4+Y)
(3/4+Z,3/4+Y,1/4-X)
(1/4+Z,1/4-Y,1/4+X)
(3/4-Z,1/4+Y,3/4+X)
(1/4-Z,3/4-Y,3/4-X)
(1/2+X,Y,1/2+Z)
(1/2-X,1/2-Y,Z)
(-X,1/2+Y,1/2-Z)
(X,-Y,-Z)
(1/2+Z,X,1/2+Y)
(Z,-X,-Y)
(1/2-Z,1/2-X,Y)
(-Z,1/2+X,1/2-Y)
(1/2+Y,Z,1/2+X)
(-Y,1/2+Z,1/2-X)
(Y,-Z,-X)
(1/2-Y,1/2-Z,X)
(1/4+Y,1/4+X,1/4-Z)
(3/4-Y,1/4-X,3/4-Z)
(3/4+Y,3/4-X,1/4+Z)
(1/4-Y,3/4+X,3/4+Z)
(1/4+X,1/4+Z,1/4-Y)
(1/4-X,3/4+Z,3/4+Y)
(3/4-X,1/4-Z,3/4-Y)
(3/4+X,3/4-Z,1/4+Y)
(1/4+Z,1/4+Y,1/4-X)
(3/4+Z,3/4-Y,1/4+X)
(1/4-Z,3/4+Y,3/4+X)
(3/4-Z,1/4-Y,3/4-X)
(1/2+X,1/2+Y,Z)
(1/2-X,-Y,1/2+Z)
(-X,Y,-Z)
(X,1/2-Y,1/2-Z)
(1/2+Z,1/2+X,Y)
(Z,1/2-X,1/2-Y)
(1/2-Z,-X,1/2+Y)
(-Z,X,-Y)
(1/2+Y,1/2+Z,X)
(-Y,Z,-X)
(Y,1/2-Z,1/2-X)
(1/2-Y,-Z,1/2+X)
(1/4+Y,3/4+X,3/4-Z)
(3/4-Y,3/4-X,1/4-Z)
(3/4+Y,1/4-X,3/4+Z)
(1/4-Y,1/4+X,1/4+Z)
(1/4+X,3/4+Z,3/4-Y)
(1/4-X,1/4+Z,1/4+Y)
(3/4-X,3/4-Z,1/4-Y)
(3/4+X,1/4-Z,3/4+Y)
(1/4+Z,3/4+Y,3/4-X)
(3/4+Z,1/4-Y,3/4+X)
(1/4-Z,1/4+Y,1/4+X)
(3/4-Z,3/4-Y,1/4-X)

I 4 3 2	48
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(Y,X,-Z)
(-Y,-X,-Z)
(Y,-X,Z)
(-Y,X,Z)
(X,Z,-Y)
(-X,Z,Y)
(-X,-Z,-Y)
(X,-Z,Y)
(Z,Y,-X)
(Z,-Y,X)
(-Z,Y,X)
(-Z,-Y,-X)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2-Z)
(1/2+Z,1/2+X,1/2+Y)
(1/2+Z,1/2-X,1/2-Y)
(1/2-Z,1/2-X,1/2+Y)
(1/2-Z,1/2+X,1/2-Y)
(1/2+Y,1/2+Z,1/2+X)
(1/2-Y,1/2+Z,1/2-X)
(1/2+Y,1/2-Z,1/2-X)
(1/2-Y,1/2-Z,1/2+X)
(1/2+Y,1/2+X,1/2-Z)
(1/2-Y,1/2-X,1/2-Z)
(1/2+Y,1/2-X,1/2+Z)
(1/2-Y,1/2+X,1/2+Z)
(1/2+X,1/2+Z,1/2-Y)
(1/2-X,1/2+Z,1/2+Y)
(1/2-X,1/2-Z,1/2-Y)
(1/2+X,1/2-Z,1/2+Y)
(1/2+Z,1/2+Y,1/2-X)
(1/2+Z,1/2-Y,1/2+X)
(1/2-Z,1/2+Y,1/2+X)
(1/2-Z,1/2-Y,1/2-X)

P 43 3 2	24
(X,Y,Z)
(1/2-X,-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,-Z)
(Z,X,Y)
(1/2+Z,1/2-X,-Y)
(1/2-Z,-X,1/2+Y)
(-Z,1/2+X,1/2-Y)
(Y,Z,X)
(-Y,1/2+Z,1/2-X)
(1/2+Y,1/2-Z,-X)
(1/2-Y,-Z,1/2+X)
(1/4+Y,3/4+X,3/4-Z)
(1/4-Y,1/4-X,1/4-Z)
(3/4+Y,3/4-X,1/4+Z)
(3/4-Y,1/4+X,3/4+Z)
(1/4+X,3/4+Z,3/4-Y)
(3/4-X,1/4+Z,3/4+Y)
(1/4-X,1/4-Z,1/4-Y)
(3/4+X,3/4-Z,1/4+Y)
(1/4+Z,3/4+Y,3/4-X)
(3/4+Z,3/4-Y,1/4+X)
(3/4-Z,1/4+Y,3/4+X)
(1/4-Z,1/4-Y,1/4-X)

P 41 3 2	24
(X,Y,Z)
(1/2-X,-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,-Z)
(Z,X,Y)
(1/2+Z,1/2-X,-Y)
(1/2-Z,-X,1/2+Y)
(-Z,1/2+X,1/2-Y)
(Y,Z,X)
(-Y,1/2+Z,1/2-X)
(1/2+Y,1/2-Z,-X)
(1/2-Y,-Z,1/2+X)
(3/4+Y,1/4+X,1/4-Z)
(3/4-Y,3/4-X,3/4-Z)
(1/4+Y,1/4-X,3/4+Z)
(1/4-Y,3/4+X,1/4+Z)
(3/4+X,1/4+Z,1/4-Y)
(1/4-X,3/4+Z,1/4+Y)
(3/4-X,3/4-Z,3/4-Y)
(1/4+X,1/4-Z,3/4+Y)
(3/4+Z,1/4+Y,1/4-X)
(1/4+Z,1/4-Y,3/4+X)
(1/4-Z,3/4+Y,1/4+X)
(3/4-Z,3/4-Y,3/4-X)

I 41 3 2	48
(X,Y,Z)
(1/2-X,-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,-Z)
(Z,X,Y)
(1/2+Z,1/2-X,-Y)
(1/2-Z,-X,1/2+Y)
(-Z,1/2+X,1/2-Y)
(Y,Z,X)
(-Y,1/2+Z,1/2-X)
(1/2+Y,1/2-Z,-X)
(1/2-Y,-Z,1/2+X)
(3/4+Y,1/4+X,1/4-Z)
(3/4-Y,3/4-X,3/4-Z)
(1/4+Y,1/4-X,3/4+Z)
(1/4-Y,3/4+X,1/4+Z)
(3/4+X,1/4+Z,1/4-Y)
(1/4-X,3/4+Z,1/4+Y)
(3/4-X,3/4-Z,3/4-Y)
(1/4+X,1/4-Z,3/4+Y)
(3/4+Z,1/4+Y,1/4-X)
(1/4+Z,1/4-Y,3/4+X)
(1/4-Z,3/4+Y,1/4+X)
(3/4-Z,3/4-Y,3/4-X)
(1/2+X,1/2+Y,1/2+Z)
(-X,1/2-Y,Z)
(1/2-X,Y,-Z)
(X,-Y,1/2-Z)
(1/2+Z,1/2+X,1/2+Y)
(Z,-X,1/2-Y)
(-Z,1/2-X,Y)
(1/2-Z,X,-Y)
(1/2+Y,1/2+Z,1/2+X)
(1/2-Y,Z,-X)
(Y,-Z,1/2-X)
(-Y,1/2-Z,X)
(1/4+Y,3/4+X,3/4-Z)
(1/4-Y,1/4-X,1/4-Z)
(3/4+Y,3/4-X,1/4+Z)
(3/4-Y,1/4+X,3/4+Z)
(1/4+X,3/4+Z,3/4-Y)
(3/4-X,1/4+Z,3/4+Y)
(1/4-X,1/4-Z,1/4-Y)
(3/4+X,3/4-Z,1/4+Y)
(1/4+Z,3/4+Y,3/4-X)
(3/4+Z,3/4-Y,1/4+X)
(3/4-Z,1/4+Y,3/4+X)
(1/4-Z,1/4-Y,1/4-X)

P -4 3 M	24
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(Y,X,Z)
(-Y,-X,Z)
(Y,-X,-Z)
(-Y,X,-Z)
(X,Z,Y)
(-X,Z,-Y)
(-X,-Z,Y)
(X,-Z,-Y)
(Z,Y,X)
(Z,-Y,-X)
(-Z,Y,-X)
(-Z,-Y,X)

F 4 -3 M	96
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(Y,X,Z)
(-Y,-X,Z)
(Y,-X,-Z)
(-Y,X,-Z)
(X,Z,Y)
(-X,Z,-Y)
(-X,-Z,Y)
(X,-Z,-Y)
(Z,Y,X)
(Z,-Y,-X)
(-Z,Y,-X)
(-Z,-Y,X)
(X,1/2+Y,1/2+Z)
(-X,1/2-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(X,1/2-Y,1/2-Z)
(Z,1/2+X,1/2+Y)
(Z,1/2-X,1/2-Y)
(-Z,1/2-X,1/2+Y)
(-Z,1/2+X,1/2-Y)
(Y,1/2+Z,1/2+X)
(-Y,1/2+Z,1/2-X)
(Y,1/2-Z,1/2-X)
(-Y,1/2-Z,1/2+X)
(Y,1/2+X,1/2+Z)
(-Y,1/2-X,1/2+Z)
(Y,1/2-X,1/2-Z)
(-Y,1/2+X,1/2-Z)
(X,1/2+Z,1/2+Y)
(-X,1/2+Z,1/2-Y)
(-X,1/2-Z,1/2+Y)
(X,1/2-Z,1/2-Y)
(Z,1/2+Y,1/2+X)
(Z,1/2-Y,1/2-X)
(-Z,1/2+Y,1/2-X)
(-Z,1/2-Y,1/2+X)
(1/2+X,Y,1/2+Z)
(1/2-X,-Y,1/2+Z)
(1/2-X,Y,1/2-Z)
(1/2+X,-Y,1/2-Z)
(1/2+Z,X,1/2+Y)
(1/2+Z,-X,1/2-Y)
(1/2-Z,-X,1/2+Y)
(1/2-Z,X,1/2-Y)
(1/2+Y,Z,1/2+X)
(1/2-Y,Z,1/2-X)
(1/2+Y,-Z,1/2-X)
(1/2-Y,-Z,1/2+X)
(1/2+Y,X,1/2+Z)
(1/2-Y,-X,1/2+Z)
(1/2+Y,-X,1/2-Z)
(1/2-Y,X,1/2-Z)
(1/2+X,Z,1/2+Y)
(1/2-X,Z,1/2-Y)
(1/2-X,-Z,1/2+Y)
(1/2+X,-Z,1/2-Y)
(1/2+Z,Y,1/2+X)
(1/2+Z,-Y,1/2-X)
(1/2-Z,Y,1/2-X)
(1/2-Z,-Y,1/2+X)
(1/2+X,1/2+Y,Z)
(1/2-X,1/2-Y,Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,1/2-Y,-Z)
(1/2+Z,1/2+X,Y)
(1/2+Z,1/2-X,-Y)
(1/2-Z,1/2-X,Y)
(1/2-Z,1/2+X,-Y)
(1/2+Y,1/2+Z,X)
(1/2-Y,1/2+Z,-X)
(1/2+Y,1/2-Z,-X)
(1/2-Y,1/2-Z,X)
(1/2+Y,1/2+X,Z)
(1/2-Y,1/2-X,Z)
(1/2+Y,1/2-X,-Z)
(1/2-Y,1/2+X,-Z)
(1/2+X,1/2+Z,Y)
(1/2-X,1/2+Z,-Y)
(1/2-X,1/2-Z,Y)
(1/2+X,1/2-Z,-Y)
(1/2+Z,1/2+Y,X)
(1/2+Z,1/2-Y,-X)
(1/2-Z,1/2+Y,-X)
(1/2-Z,1/2-Y,X)

I -4 3 M	48
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(Y,X,Z)
(-Y,-X,Z)
(Y,-X,-Z)
(-Y,X,-Z)
(X,Z,Y)
(-X,Z,-Y)
(-X,-Z,Y)
(X,-Z,-Y)
(Z,Y,X)
(Z,-Y,-X)
(-Z,Y,-X)
(-Z,-Y,X)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2-Z)
(1/2+Z,1/2+X,1/2+Y)
(1/2+Z,1/2-X,1/2-Y)
(1/2-Z,1/2-X,1/2+Y)
(1/2-Z,1/2+X,1/2-Y)
(1/2+Y,1/2+Z,1/2+X)
(1/2-Y,1/2+Z,1/2-X)
(1/2+Y,1/2-Z,1/2-X)
(1/2-Y,1/2-Z,1/2+X)
(1/2+Y,1/2+X,1/2+Z)
(1/2-Y,1/2-X,1/2+Z)
(1/2+Y,1/2-X,1/2-Z)
(1/2-Y,1/2+X,1/2-Z)
(1/2+X,1/2+Z,1/2+Y)
(1/2-X,1/2+Z,1/2-Y)
(1/2-X,1/2-Z,1/2+Y)
(1/2+X,1/2-Z,1/2-Y)
(1/2+Z,1/2+Y,1/2+X)
(1/2+Z,1/2-Y,1/2-X)
(1/2-Z,1/2+Y,1/2-X)
(1/2-Z,1/2-Y,1/2+X)

P -4 3 N	24
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(1/2+Y,1/2+X,1/2+Z)
(1/2-Y,1/2-X,1/2+Z)
(1/2+Y,1/2-X,1/2-Z)
(1/2-Y,1/2+X,1/2-Z)
(1/2+X,1/2+Z,1/2+Y)
(1/2-X,1/2+Z,1/2-Y)
(1/2-X,1/2-Z,1/2+Y)
(1/2+X,1/2-Z,1/2-Y)
(1/2+Z,1/2+Y,1/2+X)
(1/2+Z,1/2-Y,1/2-X)
(1/2-Z,1/2+Y,1/2-X)
(1/2-Z,1/2-Y,1/2+X)

F -4 3 C	96
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(1/2+Y,1/2+X,1/2+Z)
(1/2-Y,1/2-X,1/2+Z)
(1/2+Y,1/2-X,1/2-Z)
(1/2-Y,1/2+X,1/2-Z)
(1/2+X,1/2+Z,1/2+Y)
(1/2-X,1/2+Z,1/2-Y)
(1/2-X,1/2-Z,1/2+Y)
(1/2+X,1/2-Z,1/2-Y)
(1/2+Z,1/2+Y,1/2+X)
(1/2+Z,1/2-Y,1/2-X)
(1/2-Z,1/2+Y,1/2-X)
(1/2-Z,1/2-Y,1/2+X)
(X,Y+1/2,Z+1/2)
(-X,-Y+1/2,Z+1/2)
(-X,Y+1/2,-Z+1/2)
(X,-Y+1/2,-Z+1/2)
(Z,X+1/2,Y+1/2)
(Z,-X+1/2,-Y+1/2)
(-Z,-X+1/2,Y+1/2)
(-Z,X+1/2,-Y+1/2)
(Y,Z+1/2,X+1/2)
(-Y,Z+1/2,-X+1/2)
(Y,-Z+1/2,-X+1/2)
(-Y,-Z+1/2,X+1/2)
(1/2+Y,X,Z)
(1/2-Y,-X,Z)
(1/2+Y,-X,-Z)
(1/2-Y,+X,-Z)
(1/2+X,+Z,Y)
(1/2-X,+Z,-Y)
(1/2-X,-Z,Y)
(1/2+X,-Z,-Y)
(1/2+Z,+Y,X)
(1/2+Z,-Y,-X)
(1/2-Z,+Y,-X)
(1/2-Z,-Y,X)
(X+1/2,Y,Z+1/2)
(-X+1/2,-Y,Z+1/2)
(-X+1/2,Y,-Z+1/2)
(X+1/2,-Y,-Z+1/2)
(Z+1/2,X,Y+1/2)
(Z+1/2,-X,-Y+1/2)
(-Z+1/2,-X,Y+1/2)
(-Z+1/2,X,-Y+1/2)
(Y+1/2,Z,X+1/2)
(-Y+1/2,Z,-X+1/2)
(Y+1/2,-Z,-X+1/2)
(-Y+1/2,-Z,X+1/2)
(Y,1/2+X,Z)
(-Y,1/2-X,Z)
(Y,1/2-X,-Z)
(-Y,1/2+X,-Z)
(X,1/2+Z,Y)
(-X,1/2+Z,-Y)
(-X,1/2-Z,Y)
(X,1/2-Z,-Y)
(Z,1/2+Y,X)
(Z,1/2-Y,-X)
(-Z,1/2+Y,-X)
(-Z,1/2-Y,X)
(X+1/2,Y+1/2,Z)
(-X+1/2,-Y+1/2,Z)
(-X+1/2,Y+1/2,-Z)
(X+1/2,-Y+1/2,-Z)
(Z+1/2,X+1/2,Y)
(Z+1/2,-X+1/2,-Y)
(-Z+1/2,-X+1/2,Y)
(-Z+1/2,X+1/2,-Y)
(Y+1/2,Z+1/2,X)
(-Y+1/2,Z+1/2,-X)
(Y+1/2,-Z+1/2,-X)
(-Y+1/2,-Z+1/2,X)
(Y,X,1/2+Z)
(-Y,-X,1/2+Z)
(Y,-X,1/2-Z)
(-Y,X,1/2-Z)
(X,Z,1/2+Y)
(-X,Z,1/2-Y)
(-X,-Z,1/2+Y)
(X,-Z,1/2-Y)
(Z,Y,1/2+X)
(Z,-Y,1/2-X)
(-Z,Y,1/2-X)
(-Z,-Y,1/2+X)

I -4 3 D	48
(X,Y,Z)
(1/2-X,-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,-Z)
(Z,X,Y)
(1/2+Z,1/2-X,-Y)
(1/2-Z,-X,1/2+Y)
(-Z,1/2+X,1/2-Y)
(Y,Z,X)
(-Y,1/2+Z,1/2-X)
(1/2+Y,1/2-Z,-X)
(1/2-Y,-Z,1/2+X)
(1/4+Y,1/4+X,1/4+Z)
(1/4-Y,3/4-X,3/4+Z)
(3/4+Y,1/4-X,3/4-Z)
(3/4-Y,3/4+X,1/4-Z)
(1/4+X,1/4+Z,1/4+Y)
(3/4-X,3/4+Z,1/4-Y)
(1/4-X,3/4-Z,3/4+Y)
(3/4+X,1/4-Z,3/4-Y)
(1/4+Z,1/4+Y,1/4+X)
(3/4+Z,1/4-Y,3/4-X)
(3/4-Z,3/4+Y,1/4-X)
(1/4-Z,3/4-Y,3/4+X)
(X+1/2,Y+1/2,Z+1/2)
(-X,-Y+1/2,Z)
(-X+1/2,Y,-Z)
(X,-Y,-Z+1/2)
(Z+1/2,X+1/2,Y+1/2)
(Z,-X,-Y+1/2)
(-Z,-X+1/2,Y)
(-Z+1/2,X,-Y)
(Y+1/2,Z+1/2,X+1/2)
(-Y+1/2,Z,-X)
(Y,-Z,-X+1/2)
(-Y,-Z+1/2,X)
(3/4+Y,3/4+X,3/4+Z)
(3/4-Y,1/4-X,1/4+Z)
(1/4+Y,3/4-X,1/4-Z)
(1/4-Y,1/4+X,3/4-Z)
(3/4+X,3/4+Z,3/4+Y)
(1/4-X,1/4+Z,3/4-Y)
(3/4-X,1/4-Z,1/4+Y)
(1/4+X,3/4-Z,1/4-Y)
(3/4+Z,3/4+Y,3/4+X)
(1/4+Z,3/4-Y,1/4-X)
(1/4-Z,1/4+Y,3/4-X)
(3/4-Z,1/4-Y,1/4+X)

P M -3 M	48
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(Y,X,-Z)
(-Y,-X,-Z)
(Y,-X,Z)
(-Y,X,Z)
(X,Z,-Y)
(-X,Z,Y)
(-X,-Z,-Y)
(X,-Z,Y)
(Z,Y,-X)
(Z,-Y,X)
(-Z,Y,X)
(-Z,-Y,-X)
(-X,-Y,-Z)
(X,Y,-Z)
(X,-Y,Z)
(-X,Y,Z)
(-Z,-X,-Y)
(-Z,X,Y)
(Z,X,-Y)
(Z,-X,Y)
(-Y,-Z,-X)
(Y,-Z,X)
(-Y,Z,X)
(Y,Z,-X)
(-Y,-X,Z)
(Y,X,Z)
(-Y,X,-Z)
(Y,-X,-Z)
(-X,-Z,Y)
(X,-Z,-Y)
(X,Z,Y)
(-X,Z,-Y)
(-Z,-Y,X)
(-Z,Y,-X)
(Z,-Y,-X)
(Z,Y,X)

P N -3 N	48
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(Y,X,-Z)
(-Y,-X,-Z)
(Y,-X,Z)
(-Y,X,Z)
(X,Z,-Y)
(-X,Z,Y)
(-X,-Z,-Y)
(X,-Z,Y)
(Z,Y,-X)
(Z,-Y,X)
(-Z,Y,X)
(-Z,-Y,-X)
(-X,-Y,-Z)
(X,Y,-Z)
(X,-Y,Z)
(-X,Y,Z)
(-Z,-X,-Y)
(-Z,X,Y)
(Z,X,-Y)
(Z,-X,Y)
(-Y,-Z,-X)
(Y,-Z,X)
(-Y,Z,X)
(Y,Z,-X)
(-Y,-X,Z)
(Y,X,Z)
(-Y,X,-Z)
(Y,-X,-Z)
(-X,-Z,Y)
(X,-Z,-Y)
(X,Z,Y)
(-X,Z,-Y)
(-Z,-Y,X)
(-Z,Y,-X)
(Z,-Y,-X)
(Z,Y,X)

P M -3 N	48
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(1/2+Y,1/2+X,1/2-Z)
(1/2-Y,1/2-X,1/2-Z)
(1/2+Y,1/2-X,1/2+Z)
(1/2-Y,1/2+X,1/2+Z)
(1/2+X,1/2+Z,1/2-Y)
(1/2-X,1/2+Z,1/2+Y)
(1/2-X,1/2-Z,1/2-Y)
(1/2+X,1/2-Z,1/2+Y)
(1/2+Z,1/2+Y,1/2-X)
(1/2+Z,1/2-Y,1/2+X)
(1/2-Z,1/2+Y,1/2+X)
(1/2-Z,1/2-Y,1/2-X)
(-X,-Y,-Z)
(X,Y,-Z)
(X,-Y,Z)
(-X,Y,Z)
(-Z,-X,-Y)
(-Z,X,Y)
(Z,X,-Y)
(Z,-X,Y)
(-Y,-Z,-X)
(Y,-Z,X)
(-Y,Z,X)
(Y,Z,-X)
(1/2-Y,1/2-X,1/2+Z)
(1/2+Y,1/2+X,1/2+Z)
(1/2-Y,1/2+X,1/2-Z)
(1/2+Y,1/2-X,1/2-Z)
(1/2-X,1/2-Z,1/2+Y)
(1/2+X,1/2-Z,1/2-Y)
(1/2+X,1/2+Z,1/2+Y)
(1/2-X,1/2+Z,1/2-Y)
(1/2-Z,1/2-Y,1/2+X)
(1/2-Z,1/2+Y,1/2-X)
(1/2+Z,1/2-Y,1/2-X)
(1/2+Z,1/2+Y,1/2+X)

P N -3 M	48
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(1/2+Y,1/2+X,1/2-Z)
(1/2-Y,1/2-X,1/2-Z)
(1/2+Y,1/2-X,1/2+Z)
(1/2-Y,1/2+X,1/2+Z)
(1/2+X,1/2+Z,1/2-Y)
(1/2-X,1/2+Z,1/2+Y)
(1/2-X,1/2-Z,1/2-Y)
(1/2+X,1/2-Z,1/2+Y)
(1/2+Z,1/2+Y,1/2-X)
(1/2+Z,1/2-Y,1/2+X)
(1/2-Z,1/2+Y,1/2+X)
(1/2-Z,1/2-Y,1/2-X)
(1/2-X,1/2-Y,1/2-Z)
(1/2+X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2+Z)
(1/2-Z,1/2-X,1/2-Y)
(1/2-Z,1/2+X,1/2+Y)
(1/2+Z,1/2+X,1/2-Y)
(1/2+Z,1/2-X,1/2+Y)
(1/2-Y,1/2-Z,1/2-X)
(1/2+Y,1/2-Z,1/2+X)
(1/2-Y,1/2+Z,1/2+X)
(1/2+Y,1/2+Z,1/2-X)
(-Y,-X,Z)
(Y,X,Z)
(-Y,X,-Z)
(Y,-X,-Z)
(-X,-Z,Y)
(X,-Z,-Y)
(X,Z,Y)
(-X,Z,-Y)
(-Z,-Y,X)
(-Z,Y,-X)
(Z,-Y,-X)
(Z,Y,X)

F M -3 M	192
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(Y,X,-Z)
(-Y,-X,-Z)
(Y,-X,Z)
(-Y,X,Z)
(X,Z,-Y)
(-X,Z,Y)
(-X,-Z,-Y)
(X,-Z,Y)
(Z,Y,-X)
(Z,-Y,X)
(-Z,Y,X)
(-Z,-Y,-X)
(-X,-Y,-Z)
(X,Y,-Z)
(X,-Y,Z)
(-X,Y,Z)
(-Z,-X,-Y)
(-Z,X,Y)
(Z,X,-Y)
(Z,-X,Y)
(-Y,-Z,-X)
(Y,-Z,X)
(-Y,Z,X)
(Y,Z,-X)
(-Y,-X,Z)
(Y,X,Z)
(-Y,X,-Z)
(Y,-X,-Z)
(-X,-Z,Y)
(X,-Z,-Y)
(X,Z,Y)
(-X,Z,-Y)
(-Z,-Y,X)
(-Z,Y,-X)
(Z,-Y,-X)
(Z,Y,X)
(X,1/2+Y,1/2+Z)
(-X,1/2-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(X,1/2-Y,1/2-Z)
(Z,1/2+X,1/2+Y)
(Z,1/2-X,1/2-Y)
(-Z,1/2-X,1/2+Y)
(-Z,1/2+X,1/2-Y)
(Y,1/2+Z,1/2+X)
(-Y,1/2+Z,1/2-X)
(Y,1/2-Z,1/2-X)
(-Y,1/2-Z,1/2+X)
(Y,1/2+X,1/2-Z)
(-Y,1/2-X,1/2-Z)
(Y,1/2-X,1/2+Z)
(-Y,1/2+X,1/2+Z)
(X,1/2+Z,1/2-Y)
(-X,1/2+Z,1/2+Y)
(-X,1/2-Z,1/2-Y)
(X,1/2-Z,1/2+Y)
(Z,1/2+Y,1/2-X)
(Z,1/2-Y,1/2+X)
(-Z,1/2+Y,1/2+X)
(-Z,1/2-Y,1/2-X)
(-X,1/2-Y,1/2-Z)
(X,1/2+Y,1/2-Z)
(X,1/2-Y,1/2+Z)
(-X,1/2+Y,1/2+Z)
(-Z,1/2-X,1/2-Y)
(-Z,1/2+X,1/2+Y)
(Z,1/2+X,1/2-Y)
(Z,1/2-X,1/2+Y)
(-Y,1/2-Z,1/2-X)
(Y,1/2-Z,1/2+X)
(-Y,1/2+Z,1/2+X)
(Y,1/2+Z,1/2-X)
(-Y,1/2-X,1/2+Z)
(Y,1/2+X,1/2+Z)
(-Y,1/2+X,1/2-Z)
(Y,1/2-X,1/2-Z)
(-X,1/2-Z,1/2+Y)
(X,1/2-Z,1/2-Y)
(X,1/2+Z,1/2+Y)
(-X,1/2+Z,1/2-Y)
(-Z,1/2-Y,1/2+X)
(-Z,1/2+Y,1/2-X)
(Z,1/2-Y,1/2-X)
(Z,1/2+Y,1/2+X)
(1/2+X,Y,1/2+Z)
(1/2-X,-Y,1/2+Z)
(1/2-X,Y,1/2-Z)
(1/2+X,-Y,1/2-Z)
(1/2+Z,X,1/2+Y)
(1/2+Z,-X,1/2-Y)
(1/2-Z,-X,1/2+Y)
(1/2-Z,X,1/2-Y)
(1/2+Y,Z,1/2+X)
(1/2-Y,Z,1/2-X)
(1/2+Y,-Z,1/2-X)
(1/2-Y,-Z,1/2+X)
(1/2+Y,X,1/2-Z)
(1/2-Y,-X,1/2-Z)
(1/2+Y,-X,1/2+Z)
(1/2-Y,X,1/2+Z)
(1/2+X,Z,1/2-Y)
(1/2-X,Z,1/2+Y)
(1/2-X,-Z,1/2-Y)
(1/2+X,-Z,1/2+Y)
(1/2+Z,Y,1/2-X)
(1/2+Z,-Y,1/2+X)
(1/2-Z,Y,1/2+X)
(1/2-Z,-Y,1/2-X)
(1/2-X,-Y,1/2-Z)
(1/2+X,Y,1/2-Z)
(1/2+X,-Y,1/2+Z)
(1/2-X,Y,1/2+Z)
(1/2-Z,-X,1/2-Y)
(1/2-Z,X,1/2+Y)
(1/2+Z,X,1/2-Y)
(1/2+Z,-X,1/2+Y)
(1/2-Y,-Z,1/2-X)
(1/2+Y,-Z,1/2+X)
(1/2-Y,Z,1/2+X)
(1/2+Y,Z,1/2-X)
(1/2-Y,-X,1/2+Z)
(1/2+Y,X,1/2+Z)
(1/2-Y,X,1/2-Z)
(1/2+Y,-X,1/2-Z)
(1/2-X,-Z,1/2+Y)
(1/2+X,-Z,1/2-Y)
(1/2+X,Z,1/2+Y)
(1/2-X,Z,1/2-Y)
(1/2-Z,-Y,1/2+X)
(1/2-Z,Y,1/2-X)
(1/2+Z,-Y,1/2-X)
(1/2+Z,Y,1/2+X)
(1/2+X,1/2+Y,Z)
(1/2-X,1/2-Y,Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,1/2-Y,-Z)
(1/2+Z,1/2+X,Y)
(1/2+Z,1/2-X,-Y)
(1/2-Z,1/2-X,Y)
(1/2-Z,1/2+X,-Y)
(1/2+Y,1/2+Z,X)
(1/2-Y,1/2+Z,-X)
(1/2+Y,1/2-Z,-X)
(1/2-Y,1/2-Z,X)
(1/2+Y,1/2+X,-Z)
(1/2-Y,1/2-X,-Z)
(1/2+Y,1/2-X,Z)
(1/2-Y,1/2+X,Z)
(1/2+X,1/2+Z,-Y)
(1/2-X,1/2+Z,Y)
(1/2-X,1/2-Z,-Y)
(1/2+X,1/2-Z,Y)
(1/2+Z,1/2+Y,-X)
(1/2+Z,1/2-Y,X)
(1/2-Z,1/2+Y,X)
(1/2-Z,1/2-Y,-X)
(1/2-X,1/2-Y,-Z)
(1/2+X,1/2+Y,-Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,Z)
(1/2-Z,1/2-X,-Y)
(1/2-Z,1/2+X,Y)
(1/2+Z,1/2+X,-Y)
(1/2+Z,1/2-X,Y)
(1/2-Y,1/2-Z,-X)
(1/2+Y,1/2-Z,X)
(1/2-Y,1/2+Z,X)
(1/2+Y,1/2+Z,-X)
(1/2-Y,1/2-X,Z)
(1/2+Y,1/2+X,Z)
(1/2-Y,1/2+X,-Z)
(1/2+Y,1/2-X,-Z)
(1/2-X,1/2-Z,Y)
(1/2+X,1/2-Z,-Y)
(1/2+X,1/2+Z,Y)
(1/2-X,1/2+Z,-Y)
(1/2-Z,1/2-Y,X)
(1/2-Z,1/2+Y,-X)
(1/2+Z,1/2-Y,-X)
(1/2+Z,1/2+Y,X)

F M -3 C	192
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(1/2+Y,1/2+X,1/2-Z)
(1/2-Y,1/2-X,1/2-Z)
(1/2+Y,1/2-X,1/2+Z)
(1/2-Y,1/2+X,1/2+Z)
(1/2+X,1/2+Z,1/2-Y)
(1/2-X,1/2+Z,1/2+Y)
(1/2-X,1/2-Z,1/2-Y)
(1/2+X,1/2-Z,1/2+Y)
(1/2+Z,1/2+Y,1/2-X)
(1/2+Z,1/2-Y,1/2+X)
(1/2-Z,1/2+Y,1/2+X)
(1/2-Z,1/2-Y,1/2-X)
(-X,-Y,-Z)
(X,Y,-Z)
(X,-Y,Z)
(-X,Y,Z)
(-Z,-X,-Y)
(-Z,X,Y)
(Z,X,-Y)
(Z,-X,Y)
(-Y,-Z,-X)
(Y,-Z,X)
(-Y,Z,X)
(Y,Z,-X)
(1/2-Y,1/2-X,1/2+Z)
(1/2+Y,1/2+X,1/2+Z)
(1/2-Y,1/2+X,1/2-Z)
(1/2+Y,1/2-X,1/2-Z)
(1/2-X,1/2-Z,1/2+Y)
(1/2+X,1/2-Z,1/2-Y)
(1/2+X,1/2+Z,1/2+Y)
(1/2-X,1/2+Z,1/2-Y)
(1/2-Z,1/2-Y,1/2+X)
(1/2-Z,1/2+Y,1/2-X)
(1/2+Z,1/2-Y,1/2-X)
(1/2+Z,1/2+Y,1/2+X)
(X,1/2+Y,1/2+Z)
(-X,1/2-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(X,1/2-Y,1/2-Z)
(Z,1/2+X,1/2+Y)
(Z,1/2-X,1/2-Y)
(-Z,1/2-X,1/2+Y)
(-Z,1/2+X,1/2-Y)
(Y,1/2+Z,1/2+X)
(-Y,1/2+Z,1/2-X)
(Y,1/2-Z,1/2-X)
(-Y,1/2-Z,1/2+X)
(1/2+Y,X,-Z)
(1/2-Y,-X,-Z)
(1/2+Y,-X,Z)
(1/2-Y,X,Z)
(1/2+X,Z,-Y)
(1/2-X,Z,Y)
(1/2-X,-Z,-Y)
(1/2+X,-Z,Y)
(1/2+Z,Y,-X)
(1/2+Z,-Y,X)
(1/2-Z,Y,X)
(1/2-Z,-Y,-X)
(-X,1/2-Y,1/2-Z)
(X,1/2+Y,1/2-Z)
(X,1/2-Y,1/2+Z)
(-X,1/2+Y,1/2+Z)
(-Z,1/2-X,1/2-Y)
(-Z,1/2+X,1/2+Y)
(Z,1/2+X,1/2-Y)
(Z,1/2-X,1/2+Y)
(-Y,1/2-Z,1/2-X)
(Y,1/2-Z,1/2+X)
(-Y,1/2+Z,1/2+X)
(Y,1/2+Z,1/2-X)
(1/2-Y,-X,Z)
(1/2+Y,X,Z)
(1/2-Y,X,-Z)
(1/2+Y,-X,-Z)
(1/2-X,-Z,Y)
(1/2+X,-Z,-Y)
(1/2+X,Z,Y)
(1/2-X,Z,-Y)
(1/2-Z,-Y,X)
(1/2-Z,Y,-X)
(1/2+Z,-Y,-X)
(1/2+Z,Y,X)
(1/2+X,Y,1/2+Z)
(1/2-X,-Y,1/2+Z)
(1/2-X,Y,1/2-Z)
(1/2+X,-Y,1/2-Z)
(1/2+Z,X,1/2+Y)
(1/2+Z,-X,1/2-Y)
(1/2-Z,-X,1/2+Y)
(1/2-Z,X,1/2-Y)
(1/2+Y,Z,1/2+X)
(1/2-Y,Z,1/2-X)
(1/2+Y,-Z,1/2-X)
(1/2-Y,-Z,1/2+X)
(Y,1/2+X,-Z)
(-Y,1/2-X,-Z)
(Y,1/2-X,Z)
(-Y,1/2+X,Z)
(X,1/2+Z,-Y)
(-X,1/2+Z,Y)
(-X,1/2-Z,-Y)
(X,1/2-Z,Y)
(Z,1/2+Y,-X)
(Z,1/2-Y,X)
(-Z,1/2+Y,X)
(-Z,1/2-Y,-X)
(1/2-X,-Y,1/2-Z)
(1/2+X,Y,1/2-Z)
(1/2+X,-Y,1/2+Z)
(1/2-X,Y,1/2+Z)
(1/2-Z,-X,1/2-Y)
(1/2-Z,X,1/2+Y)
(1/2+Z,X,1/2-Y)
(1/2+Z,-X,1/2+Y)
(1/2-Y,-Z,1/2-X)
(1/2+Y,-Z,1/2+X)
(1/2-Y,Z,1/2+X)
(1/2+Y,Z,1/2-X)
(-Y,1/2-X,Z)
(Y,1/2+X,Z)
(-Y,1/2+X,-Z)
(Y,1/2-X,-Z)
(-X,1/2-Z,Y)
(X,1/2-Z,-Y)
(X,1/2+Z,Y)
(-X,1/2+Z,-Y)
(-Z,1/2-Y,X)
(-Z,1/2+Y,-X)
(Z,1/2-Y,-X)
(Z,1/2+Y,X)
(1/2+X,1/2+Y,Z)
(1/2-X,1/2-Y,Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,1/2-Y,-Z)
(1/2+Z,1/2+X,Y)
(1/2+Z,1/2-X,-Y)
(1/2-Z,1/2-X,Y)
(1/2-Z,1/2+X,-Y)
(1/2+Y,1/2+Z,X)
(1/2-Y,1/2+Z,-X)
(1/2+Y,1/2-Z,-X)
(1/2-Y,1/2-Z,X)
(Y,X,1/2-Z)
(-Y,-X,1/2-Z)
(Y,-X,1/2+Z)
(-Y,X,1/2+Z)
(X,Z,1/2-Y)
(-X,Z,1/2+Y)
(-X,-Z,1/2-Y)
(X,-Z,1/2+Y)
(Z,Y,1/2-X)
(Z,-Y,1/2+X)
(-Z,Y,1/2+X)
(-Z,-Y,1/2-X)
(1/2-X,1/2-Y,-Z)
(1/2+X,1/2+Y,-Z)
(1/2+X,1/2-Y,Z)
(1/2-X,1/2+Y,Z)
(1/2-Z,1/2-X,-Y)
(1/2-Z,1/2+X,Y)
(1/2+Z,1/2+X,-Y)
(1/2+Z,1/2-X,Y)
(1/2-Y,1/2-Z,-X)
(1/2+Y,1/2-Z,X)
(1/2-Y,1/2+Z,X)
(1/2+Y,1/2+Z,-X)
(-Y,-X,1/2+Z)
(Y,X,1/2+Z)
(-Y,X,1/2-Z)
(Y,-X,1/2-Z)
(-X,-Z,1/2+Y)
(X,-Z,1/2-Y)
(X,Z,1/2+Y)
(-X,Z,1/2-Y)
(-Z,-Y,1/2+X)
(-Z,Y,1/2-X)
(Z,-Y,1/2-X)
(Z,Y,1/2+X)

F D -3 M	192
(X,Y,Z)
(-X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,-Y,1/2-Z)
(Z,X,Y)
(1/2+Z,-X,1/2-Y)
(-Z,1/2-X,1/2+Y)
(1/2-Z,1/2+X,-Y)
(Y,Z,X)
(1/2-Y,1/2+Z,-X)
(1/2+Y,-Z,1/2-X)
(-Y,1/2-Z,1/2+X)
(3/4+Y,1/4+X,3/4-Z)
(1/4-Y,1/4-X,1/4-Z)
(1/4+Y,3/4-X,3/4+Z)
(3/4-Y,3/4+X,1/4+Z)
(3/4+X,1/4+Z,3/4-Y)
(3/4-X,3/4+Z,1/4+Y)
(1/4-X,1/4-Z,1/4-Y)
(1/4+X,3/4-Z,3/4+Y)
(3/4+Z,1/4+Y,3/4-X)
(1/4+Z,3/4-Y,3/4+X)
(3/4-Z,3/4+Y,1/4+X)
(1/4-Z,1/4-Y,1/4-X)
(1/4-X,1/4-Y,1/4-Z)
(1/4+X,3/4+Y,3/4-Z)
(3/4+X,3/4-Y,1/4+Z)
(3/4-X,1/4+Y,3/4+Z)
(1/4-Z,1/4-X,1/4-Y)
(3/4-Z,1/4+X,3/4+Y)
(1/4+Z,3/4+X,3/4-Y)
(3/4+Z,3/4-X,1/4+Y)
(1/4-Y,1/4-Z,1/4-X)
(3/4+Y,3/4-Z,1/4+X)
(3/4-Y,1/4+Z,3/4+X)
(1/4+Y,3/4+Z,3/4-X)
(1/2-Y,-X,1/2+Z)
(Y,X,Z)
(-Y,1/2+X,1/2-Z)
(1/2+Y,1/2-X,-Z)
(1/2-X,-Z,1/2+Y)
(1/2+X,1/2-Z,-Y)
(X,Z,Y)
(-X,1/2+Z,1/2-Y)
(1/2-Z,-Y,1/2+X)
(-Z,1/2+Y,1/2-X)
(1/2+Z,1/2-Y,-X)
(Z,Y,X)
(X,1/2+Y,1/2+Z)
(-X,-Y,Z)
(1/2-X,Y,1/2-Z)
(1/2+X,1/2-Y,-Z)
(Z,1/2+X,1/2+Y)
(1/2+Z,1/2-X,-Y)
(-Z,-X,Y)
(1/2-Z,X,1/2-Y)
(Y,1/2+Z,1/2+X)
(1/2-Y,Z,1/2-X)
(1/2+Y,1/2-Z,-X)
(-Y,-Z,X)
(3/4+Y,3/4+X,1/4-Z)
(1/4-Y,3/4-X,3/4-Z)
(1/4+Y,1/4-X,1/4+Z)
(3/4-Y,1/4+X,3/4+Z)
(3/4+X,3/4+Z,1/4-Y)
(3/4-X,1/4+Z,3/4+Y)
(1/4-X,3/4-Z,3/4-Y)
(1/4+X,1/4-Z,1/4+Y)
(3/4+Z,3/4+Y,1/4-X)
(1/4+Z,1/4-Y,1/4+X)
(3/4-Z,1/4+Y,3/4+X)
(1/4-Z,3/4-Y,3/4-X)
(1/4-X,3/4-Y,3/4-Z)
(1/4+X,1/4+Y,1/4-Z)
(3/4+X,1/4-Y,3/4+Z)
(3/4-X,3/4+Y,1/4+Z)
(1/4-Z,3/4-X,3/4-Y)
(3/4-Z,3/4+X,1/4+Y)
(1/4+Z,1/4+X,1/4-Y)
(3/4+Z,1/4-X,3/4+Y)
(1/4-Y,3/4-Z,3/4-X)
(3/4+Y,1/4-Z,3/4+X)
(3/4-Y,3/4+Z,1/4+X)
(1/4+Y,1/4+Z,1/4-X)
(1/2-Y,1/2-X,Z)
(Y,1/2+X,1/2+Z)
(-Y,X,-Z)
(1/2+Y,-X,1/2-Z)
(1/2-X,1/2-Z,Y)
(1/2+X,-Z,1/2-Y)
(X,1/2+Z,1/2+Y)
(-X,Z,-Y)
(1/2-Z,1/2-Y,X)
(-Z,Y,-X)
(1/2+Z,-Y,1/2-X)
(Z,1/2+Y,1/2+X)
(1/2+X,Y,1/2+Z)
(1/2-X,1/2-Y,Z)
(-X,1/2+Y,1/2-Z)
(X,-Y,-Z)
(1/2+Z,X,1/2+Y)
(Z,-X,-Y)
(1/2-Z,1/2-X,Y)
(-Z,1/2+X,1/2-Y)
(1/2+Y,Z,1/2+X)
(-Y,1/2+Z,1/2-X)
(Y,-Z,-X)
(1/2-Y,1/2-Z,X)
(1/4+Y,1/4+X,1/4-Z)
(3/4-Y,1/4-X,3/4-Z)
(3/4+Y,3/4-X,1/4+Z)
(1/4-Y,3/4+X,3/4+Z)
(1/4+X,1/4+Z,1/4-Y)
(1/4-X,3/4+Z,3/4+Y)
(3/4-X,1/4-Z,3/4-Y)
(3/4+X,3/4-Z,1/4+Y)
(1/4+Z,1/4+Y,1/4-X)
(3/4+Z,3/4-Y,1/4+X)
(1/4-Z,3/4+Y,3/4+X)
(3/4-Z,1/4-Y,3/4-X)
(3/4-X,1/4-Y,3/4-Z)
(3/4+X,3/4+Y,1/4-Z)
(1/4+X,3/4-Y,3/4+Z)
(1/4-X,1/4+Y,1/4+Z)
(3/4-Z,1/4-X,3/4-Y)
(1/4-Z,1/4+X,1/4+Y)
(3/4+Z,3/4+X,1/4-Y)
(1/4+Z,3/4-X,3/4+Y)
(3/4-Y,1/4-Z,3/4-X)
(1/4+Y,3/4-Z,3/4+X)
(1/4-Y,1/4+Z,1/4+X)
(3/4+Y,3/4+Z,1/4-X)
(-Y,-X,Z)
(1/2+Y,X,1/2+Z)
(1/2-Y,1/2+X,-Z)
(Y,1/2-X,1/2-Z)
(-X,-Z,Y)
(X,1/2-Z,1/2-Y)
(1/2+X,Z,1/2+Y)
(1/2-X,1/2+Z,-Y)
(-Z,-Y,X)
(1/2-Z,1/2+Y,-X)
(Z,1/2-Y,1/2-X)
(1/2+Z,Y,1/2+X)
(1/2+X,1/2+Y,Z)
(1/2-X,-Y,1/2+Z)
(-X,Y,-Z)
(X,1/2-Y,1/2-Z)
(1/2+Z,1/2+X,Y)
(Z,1/2-X,1/2-Y)
(1/2-Z,-X,1/2+Y)
(-Z,X,-Y)
(1/2+Y,1/2+Z,X)
(-Y,Z,-X)
(Y,1/2-Z,1/2-X)
(1/2-Y,-Z,1/2+X)
(1/4+Y,3/4+X,3/4-Z)
(3/4-Y,3/4-X,1/4-Z)
(3/4+Y,1/4-X,3/4+Z)
(1/4-Y,1/4+X,1/4+Z)
(1/4+X,3/4+Z,3/4-Y)
(1/4-X,1/4+Z,1/4+Y)
(3/4-X,3/4-Z,1/4-Y)
(3/4+X,1/4-Z,3/4+Y)
(1/4+Z,3/4+Y,3/4-X)
(3/4+Z,1/4-Y,3/4+X)
(1/4-Z,1/4+Y,1/4+X)
(3/4-Z,3/4-Y,1/4-X)
(3/4-X,3/4-Y,1/4-Z)
(3/4+X,1/4+Y,3/4-Z)
(1/4+X,1/4-Y,1/4+Z)
(1/4-X,3/4+Y,3/4+Z)
(3/4-Z,3/4-X,1/4-Y)
(1/4-Z,3/4+X,3/4+Y)
(3/4+Z,1/4+X,3/4-Y)
(1/4+Z,1/4-X,1/4+Y)
(3/4-Y,3/4-Z,1/4-X)
(1/4+Y,1/4-Z,1/4+X)
(1/4-Y,3/4+Z,3/4+X)
(3/4+Y,1/4+Z,3/4-X)
(-Y,1/2-X,1/2+Z)
(1/2+Y,1/2+X,Z)
(1/2-Y,X,1/2-Z)
(Y,-X,-Z)
(-X,1/2-Z,1/2+Y)
(X,-Z,-Y)
(1/2+X,1/2+Z,Y)
(1/2-X,Z,1/2-Y)
(-Z,1/2-Y,1/2+X)
(1/2-Z,Y,1/2-X)
(Z,-Y,-X)
(1/2+Z,1/2+Y,X)

F D -3 C	192
(X,Y,Z)
(-X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,-Z)
(1/2+X,-Y,1/2-Z)
(Z,X,Y)
(1/2+Z,-X,1/2-Y)
(-Z,1/2-X,1/2+Y)
(1/2-Z,1/2+X,-Y)
(Y,Z,X)
(1/2-Y,1/2+Z,-X)
(1/2+Y,-Z,1/2-X)
(-Y,1/2-Z,1/2+X)
(3/4+Y,1/4+X,3/4-Z)
(1/4-Y,1/4-X,1/4-Z)
(1/4+Y,3/4-X,3/4+Z)
(3/4-Y,3/4+X,1/4+Z)
(3/4+X,1/4+Z,3/4-Y)
(3/4-X,3/4+Z,1/4+Y)
(1/4-X,1/4-Z,1/4-Y)
(1/4+X,3/4-Z,3/4+Y)
(3/4+Z,1/4+Y,3/4-X)
(1/4+Z,3/4-Y,3/4+X)
(3/4-Z,3/4+Y,1/4+X)
(1/4-Z,1/4-Y,1/4-X)
(3/4-X,3/4-Y,3/4-Z)
(3/4+X,1/4+Y,1/4-Z)
(1/4+X,1/4-Y,3/4+Z)
(1/4-X,3/4+Y,1/4+Z)
(3/4-Z,3/4-X,3/4-Y)
(1/4-Z,3/4+X,1/4+Y)
(3/4+Z,1/4+X,1/4-Y)
(1/4+Z,1/4-X,3/4+Y)
(3/4-Y,3/4-Z,3/4-X)
(1/4+Y,1/4-Z,3/4+X)
(1/4-Y,3/4+Z,1/4+X)
(3/4+Y,1/4+Z,1/4-X)
(-Y,1/2-X,Z)
(1/2+Y,1/2+X,1/2+Z)
(1/2-Y,X,-Z)
(Y,-X,1/2-Z)
(-X,1/2-Z,Y)
(X,-Z,1/2-Y)
(1/2+X,1/2+Z,1/2+Y)
(1/2-X,Z,-Y)
(-Z,1/2-Y,X)
(1/2-Z,Y,-X)
(Z,-Y,1/2-X)
(1/2+Z,1/2+Y,1/2+X)
(X,1/2+Y,1/2+Z)
(-X,-Y,Z)
(1/2-X,Y,1/2-Z)
(1/2+X,1/2-Y,-Z)
(Z,1/2+X,1/2+Y)
(1/2+Z,1/2-X,-Y)
(-Z,-X,Y)
(1/2-Z,X,1/2-Y)
(Y,1/2+Z,1/2+X)
(1/2-Y,Z,1/2-X)
(1/2+Y,1/2-Z,-X)
(-Y,-Z,X)
(3/4+Y,3/4+X,1/4-Z)
(1/4-Y,3/4-X,3/4-Z)
(1/4+Y,1/4-X,1/4+Z)
(3/4-Y,1/4+X,3/4+Z)
(3/4+X,3/4+Z,1/4-Y)
(3/4-X,1/4+Z,3/4+Y)
(1/4-X,3/4-Z,3/4-Y)
(1/4+X,1/4-Z,1/4+Y)
(3/4+Z,3/4+Y,1/4-X)
(1/4+Z,1/4-Y,1/4+X)
(3/4-Z,1/4+Y,3/4+X)
(1/4-Z,3/4-Y,3/4-X)
(3/4-X,1/4-Y,1/4-Z)
(3/4+X,3/4+Y,3/4-Z)
(1/4+X,3/4-Y,1/4+Z)
(1/4-X,1/4+Y,3/4+Z)
(3/4-Z,1/4-X,1/4-Y)
(1/4-Z,1/4+X,3/4+Y)
(3/4+Z,3/4+X,3/4-Y)
(1/4+Z,3/4-X,1/4+Y)
(3/4-Y,1/4-Z,1/4-X)
(1/4+Y,3/4-Z,1/4+X)
(1/4-Y,1/4+Z,3/4+X)
(3/4+Y,3/4+Z,3/4-X)
(-Y,-X,1/2+Z)
(1/2+Y,X,Z)
(1/2-Y,1/2+X,1/2-Z)
(Y,1/2-X,-Z)
(-X,-Z,1/2+Y)
(X,1/2-Z,-Y)
(1/2+X,Z,Y)
(1/2-X,1/2+Z,1/2-Y)
(-Z,-Y,1/2+X)
(1/2-Z,1/2+Y,1/2-X)
(Z,1/2-Y,-X)
(1/2+Z,Y,X)
(1/2+X,Y,1/2+Z)
(1/2-X,1/2-Y,Z)
(-X,1/2+Y,1/2-Z)
(X,-Y,-Z)
(1/2+Z,X,1/2+Y)
(Z,-X,-Y)
(1/2-Z,1/2-X,Y)
(-Z,1/2+X,1/2-Y)
(1/2+Y,Z,1/2+X)
(-Y,1/2+Z,1/2-X)
(Y,-Z,-X)
(1/2-Y,1/2-Z,X)
(1/4+Y,1/4+X,1/4-Z)
(3/4-Y,1/4-X,3/4-Z)
(3/4+Y,3/4-X,1/4+Z)
(1/4-Y,3/4+X,3/4+Z)
(1/4+X,1/4+Z,1/4-Y)
(1/4-X,3/4+Z,3/4+Y)
(3/4-X,1/4-Z,3/4-Y)
(3/4+X,3/4-Z,1/4+Y)
(1/4+Z,1/4+Y,1/4-X)
(3/4+Z,3/4-Y,1/4+X)
(1/4-Z,3/4+Y,3/4+X)
(3/4-Z,1/4-Y,3/4-X)
(1/4-X,3/4-Y,1/4-Z)
(1/4+X,1/4+Y,3/4-Z)
(3/4+X,1/4-Y,1/4+Z)
(3/4-X,3/4+Y,3/4+Z)
(1/4-Z,3/4-X,1/4-Y)
(3/4-Z,3/4+X,3/4+Y)
(1/4+Z,1/4+X,3/4-Y)
(3/4+Z,1/4-X,1/4+Y)
(1/4-Y,3/4-Z,1/4-X)
(3/4+Y,1/4-Z,1/4+X)
(3/4-Y,3/4+Z,3/4+X)
(1/4+Y,1/4+Z,3/4-X)
(1/2-Y,1/2-X,1/2+Z)
(Y,1/2+X,Z)
(-Y,X,1/2-Z)
(1/2+Y,-X,-Z)
(1/2-X,1/2-Z,1/2+Y)
(1/2+X,-Z,-Y)
(X,1/2+Z,Y)
(-X,Z,1/2-Y)
(1/2-Z,1/2-Y,1/2+X)
(-Z,Y,1/2-X)
(1/2+Z,-Y,-X)
(Z,1/2+Y,X)
(1/2+X,1/2+Y,Z)
(1/2-X,-Y,1/2+Z)
(-X,Y,-Z)
(X,1/2-Y,1/2-Z)
(1/2+Z,1/2+X,Y)
(Z,1/2-X,1/2-Y)
(1/2-Z,-X,1/2+Y)
(-Z,X,-Y)
(1/2+Y,1/2+Z,X)
(-Y,Z,-X)
(Y,1/2-Z,1/2-X)
(1/2-Y,-Z,1/2+X)
(1/4+Y,3/4+X,3/4-Z)
(3/4-Y,3/4-X,1/4-Z)
(3/4+Y,1/4-X,3/4+Z)
(1/4-Y,1/4+X,1/4+Z)
(1/4+X,3/4+Z,3/4-Y)
(1/4-X,1/4+Z,1/4+Y)
(3/4-X,3/4-Z,1/4-Y)
(3/4+X,1/4-Z,3/4+Y)
(1/4+Z,3/4+Y,3/4-X)
(3/4+Z,1/4-Y,3/4+X)
(1/4-Z,1/4+Y,1/4+X)
(3/4-Z,3/4-Y,1/4-X)
(1/4-X,1/4-Y,3/4-Z)
(1/4+X,3/4+Y,1/4-Z)
(3/4+X,3/4-Y,3/4+Z)
(3/4-X,1/4+Y,1/4+Z)
(1/4-Z,1/4-X,3/4-Y)
(3/4-Z,1/4+X,1/4+Y)
(1/4+Z,3/4+X,1/4-Y)
(3/4+Z,3/4-X,3/4+Y)
(1/4-Y,1/4-Z,3/4-X)
(3/4+Y,3/4-Z,3/4+X)
(3/4-Y,1/4+Z,1/4+X)
(1/4+Y,3/4+Z,1/4-X)
(1/2-Y,-X,Z)
(Y,X,1/2+Z)
(-Y,1/2+X,-Z)
(1/2+Y,1/2-X,1/2-Z)
(1/2-X,-Z,Y)
(1/2+X,1/2-Z,1/2-Y)
(X,Z,1/2+Y)
(-X,1/2+Z,-Y)
(1/2-Z,-Y,X)
(-Z,1/2+Y,-X)
(1/2+Z,1/2-Y,1/2-X)
(Z,Y,1/2+X)

I M -3 M	96
(X,Y,Z)
(-X,-Y,Z)
(-X,Y,-Z)
(X,-Y,-Z)
(Z,X,Y)
(Z,-X,-Y)
(-Z,-X,Y)
(-Z,X,-Y)
(Y,Z,X)
(-Y,Z,-X)
(Y,-Z,-X)
(-Y,-Z,X)
(Y,X,-Z)
(-Y,-X,-Z)
(Y,-X,Z)
(-Y,X,Z)
(X,Z,-Y)
(-X,Z,Y)
(-X,-Z,-Y)
(X,-Z,Y)
(Z,Y,-X)
(Z,-Y,X)
(-Z,Y,X)
(-Z,-Y,-X)
(-X,-Y,-Z)
(X,Y,-Z)
(X,-Y,Z)
(-X,Y,Z)
(-Z,-X,-Y)
(-Z,X,Y)
(Z,X,-Y)
(Z,-X,Y)
(-Y,-Z,-X)
(Y,-Z,X)
(-Y,Z,X)
(Y,Z,-X)
(-Y,-X,Z)
(Y,X,Z)
(-Y,X,-Z)
(Y,-X,-Z)
(-X,-Z,Y)
(X,-Z,-Y)
(X,Z,Y)
(-X,Z,-Y)
(-Z,-Y,X)
(-Z,Y,-X)
(Z,-Y,-X)
(Z,Y,X)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2-Z)
(1/2+Z,1/2+X,1/2+Y)
(1/2+Z,1/2-X,1/2-Y)
(1/2-Z,1/2-X,1/2+Y)
(1/2-Z,1/2+X,1/2-Y)
(1/2+Y,1/2+Z,1/2+X)
(1/2-Y,1/2+Z,1/2-X)
(1/2+Y,1/2-Z,1/2-X)
(1/2-Y,1/2-Z,1/2+X)
(1/2+Y,1/2+X,1/2-Z)
(1/2-Y,1/2-X,1/2-Z)
(1/2+Y,1/2-X,1/2+Z)
(1/2-Y,1/2+X,1/2+Z)
(1/2+X,1/2+Z,1/2-Y)
(1/2-X,1/2+Z,1/2+Y)
(1/2-X,1/2-Z,1/2-Y)
(1/2+X,1/2-Z,1/2+Y)
(1/2+Z,1/2+Y,1/2-X)
(1/2+Z,1/2-Y,1/2+X)
(1/2-Z,1/2+Y,1/2+X)
(1/2-Z,1/2-Y,1/2-X)
(1/2-X,1/2-Y,1/2-Z)
(1/2+X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,1/2+Z)
(1/2-Z,1/2-X,1/2-Y)
(1/2-Z,1/2+X,1/2+Y)
(1/2+Z,1/2+X,1/2-Y)
(1/2+Z,1/2-X,1/2+Y)
(1/2-Y,1/2-Z,1/2-X)
(1/2+Y,1/2-Z,1/2+X)
(1/2-Y,1/2+Z,1/2+X)
(1/2+Y,1/2+Z,1/2-X)
(1/2-Y,1/2-X,1/2+Z)
(1/2+Y,1/2+X,1/2+Z)
(1/2-Y,1/2+X,1/2-Z)
(1/2+Y,1/2-X,1/2-Z)
(1/2-X,1/2-Z,1/2+Y)
(1/2+X,1/2-Z,1/2-Y)
(1/2+X,1/2+Z,1/2+Y)
(1/2-X,1/2+Z,1/2-Y)
(1/2-Z,1/2-Y,1/2+X)
(1/2-Z,1/2+Y,1/2-X)
(1/2+Z,1/2-Y,1/2-X)
(1/2+Z,1/2+Y,1/2+X)

I A -3 D	96
(X,Y,Z)
(1/2-X,-Y,1/2+Z)
(-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,-Z)
(Z,X,Y)
(1/2+Z,1/2-X,-Y)
(1/2-Z,-X,1/2+Y)
(-Z,1/2+X,1/2-Y)
(Y,Z,X)
(-Y,1/2+Z,1/2-X)
(1/2+Y,1/2-Z,-X)
(1/2-Y,-Z,1/2+X)
(3/4+Y,1/4+X,1/4-Z)
(3/4-Y,3/4-X,3/4-Z)
(1/4+Y,1/4-X,3/4+Z)
(1/4-Y,3/4+X,1/4+Z)
(3/4+X,1/4+Z,1/4-Y)
(1/4-X,3/4+Z,1/4+Y)
(3/4-X,3/4-Z,3/4-Y)
(1/4+X,1/4-Z,3/4+Y)
(3/4+Z,1/4+Y,1/4-X)
(1/4+Z,1/4-Y,3/4+X)
(1/4-Z,3/4+Y,1/4+X)
(3/4-Z,3/4-Y,3/4-X)
(-X,-Y,-Z)
(1/2+X,Y,1/2-Z)
(X,1/2-Y,1/2+Z)
(1/2-X,1/2+Y,Z)
(-Z,-X,-Y)
(1/2-Z,1/2+X,Y)
(1/2+Z,X,1/2-Y)
(Z,1/2-X,1/2+Y)
(-Y,-Z,-X)
(Y,1/2-Z,1/2+X)
(1/2-Y,1/2+Z,X)
(1/2+Y,Z,1/2-X)
(1/4-Y,3/4-X,3/4+Z)
(1/4+Y,1/4+X,1/4+Z)
(3/4-Y,3/4+X,1/4-Z)
(3/4+Y,1/4-X,3/4-Z)
(1/4-X,3/4-Z,3/4+Y)
(3/4+X,1/4-Z,3/4-Y)
(1/4+X,1/4+Z,1/4+Y)
(3/4-X,3/4+Z,1/4-Y)
(1/4-Z,3/4-Y,3/4+X)
(3/4-Z,3/4+Y,1/4-X)
(3/4+Z,1/4-Y,3/4-X)
(1/4+Z,1/4+Y,1/4+X)
(1/2+X,1/2+Y,1/2+Z)
(-X,1/2-Y,Z)
(1/2-X,Y,-Z)
(X,-Y,1/2-Z)
(1/2+Z,1/2+X,1/2+Y)
(Z,-X,1/2-Y)
(-Z,1/2-X,Y)
(1/2-Z,X,-Y)
(1/2+Y,1/2+Z,1/2+X)
(1/2-Y,Z,-X)
(Y,-Z,1/2-X)
(-Y,1/2-Z,X)
(1/4+Y,3/4+X,3/4-Z)
(1/4-Y,1/4-X,1/4-Z)
(3/4+Y,3/4-X,1/4+Z)
(3/4-Y,1/4+X,3/4+Z)
(1/4+X,3/4+Z,3/4-Y)
(3/4-X,1/4+Z,3/4+Y)
(1/4-X,1/4-Z,1/4-Y)
(3/4+X,3/4-Z,1/4+Y)
(1/4+Z,3/4+Y,3/4-X)
(3/4+Z,3/4-Y,1/4+X)
(3/4-Z,1/4+Y,3/4+X)
(1/4-Z,1/4-Y,1/4-X)
(1/2-X,1/2-Y,1/2-Z)
(X,1/2+Y,-Z)
(1/2+X,-Y,Z)
(-X,Y,1/2+Z)
(1/2-Z,1/2-X,1/2-Y)
(-Z,X,1/2+Y)
(Z,1/2+X,-Y)
(1/2+Z,-X,Y)
(1/2-Y,1/2-Z,1/2-X)
(1/2+Y,-Z,X)
(-Y,Z,1/2+X)
(Y,1/2+Z,-X)
(3/4-Y,1/4-X,1/4+Z)
(3/4+Y,3/4+X,3/4+Z)
(1/4-Y,1/4+X,3/4-Z)
(1/4+Y,3/4-X,1/4-Z)
(3/4-X,1/4-Z,1/4+Y)
(1/4+X,3/4-Z,1/4-Y)
(3/4+X,3/4+Z,3/4+Y)
(1/4-X,1/4+Z,3/4-Y)
(3/4-Z,1/4-Y,1/4+X)
(1/4-Z,1/4+Y,3/4-X)
(1/4+Z,3/4-Y,1/4-X)
(3/4+Z,3/4+Y,3/4+X)
"""

# -----------------------------------------------------------------------------
# The following space groups come from ccp4_symop_lib.txt on the web.
# These names are sometimes used in PDB file CRYST1 records.
#
ccp4_symop_extras = \
"""
B 2     4
(X,Y,Z)
(-X,-Y,Z)
(1/2+X,+Y,1/2+Z)
(1/2-X,-Y,1/2+Z)

P 1 1 21        2
(X,Y,Z)
(-X,-Y,1/2+Z)

P 2 21 21       4
(X,Y,Z)
(X,-Y,-Z)
(-X,1/2+Y,1/2-Z)
(-X,1/2-Y,1/2+Z)

P 21 2 21       4
(X,Y,Z)
(-X,Y,-Z)
(1/2+X,-Y,1/2-Z)
(1/2-X,-Y,1/2+Z)
"""

# -----------------------------------------------------------------------------
# Some PDB file contain CRYST1 space group names not found in the EBI list
# of standard names or the CCP4 names.  But the PDB file itself contains
# the symmetry operations, so these can be added to the table.
#
pdb_file_symop_extras = \
"""
I 1 2 1         4
(X,Y,Z)
(-X,Y,-Z)
(1/2+X,1/2+Y,1/2+Z)
(1/2-X,1/2+Y,1/2-Z)
"""

# -----------------------------------------------------------------------------
#
misc_symop_extras = \
"""
P 21/N  4
(X,Y,Z)
(-X,-Y,-Z)
(1/2-X,1/2+Y,1/2-Z)
(1/2+X,1/2-Y,1/2+Z)

R 3   3
(X,Y,Z)
(Z,X,Y)
(Y,Z,X)

R 3 2   6
(X,Y,Z)
(Z,X,Y)
(Y,Z,X)
(-Y,-X,-Z)
(-Z,-Y,-X)
(-X,-Z,-Y)
"""

# -----------------------------------------------------------------------------
#
space_group_symmetry_operations = (ebi_autodep_symmetry_operations +
                                   ccp4_symop_extras +
                                   pdb_file_symop_extras +
                                   misc_symop_extras)
